# ruff: noqa: E501
from __future__ import annotations

import io
from typing import Any

from chik_rs import serialized_length
from klvm.serialize import sexp_from_stream
from klvm.SExp import SExp
from klvm_tools import binutils

from chik.types.blockchain_format.program import INFINITE_COST, Program
from chik.types.spend_bundle import SpendBundle
from chik.util.byte_types import hexstr_to_bytes
from chik.util.ints import uint32
from chik.wallet.puzzles.load_klvm import load_klvm

TEST_GEN_DESERIALIZE = load_klvm(
    "test_generator_deserialize.clsp", package_or_requirement="chik._tests.generator.puzzles"
)
DESERIALIZE_MOD = load_klvm("chiklisp_deserialisation.clsp", package_or_requirement="chik.consensus.puzzles")

DECOMPRESS_PUZZLE = load_klvm("decompress_puzzle.clsp", package_or_requirement="chik.full_node.puzzles")
DECOMPRESS_CSE = load_klvm("decompress_coin_spend_entry.clsp", package_or_requirement="chik.full_node.puzzles")

DECOMPRESS_CSE_WITH_PREFIX = load_klvm(
    "decompress_coin_spend_entry_with_prefix.clsp", package_or_requirement="chik.full_node.puzzles"
)
DECOMPRESS_BLOCK = load_klvm("block_program_zero.clsp", package_or_requirement="chik.full_node.puzzles")
TEST_MULTIPLE = load_klvm(
    "test_multiple_generator_input_arguments.clsp", package_or_requirement="chik._tests.generator.puzzles"
)

Nil = Program.from_bytes(b"\x80")

original_generator = hexstr_to_bytes(
    "ff01ffffffa00000000000000000000000000000000000000000000000000000000000000000ff830186a080ffffff02ffff01ff02ffff01ff02ffff03ff0bffff01ff02ffff03ffff09ff05ffff1dff0bffff1effff0bff0bffff02ff06ffff04ff02ffff04ff17ff8080808080808080ffff01ff02ff17ff2f80ffff01ff088080ff0180ffff01ff04ffff04ff04ffff04ff05ffff04ffff02ff06ffff04ff02ffff04ff17ff80808080ff80808080ffff02ff17ff2f808080ff0180ffff04ffff01ff32ff02ffff03ffff07ff0580ffff01ff0bffff0102ffff02ff06ffff04ff02ffff04ff09ff80808080ffff02ff06ffff04ff02ffff04ff0dff8080808080ffff01ff0bffff0101ff058080ff0180ff018080ffff04ffff01b081963921826355dcb6c355ccf9c2637c18adf7d38ee44d803ea9ca41587e48c913d8d46896eb830aeadfc13144a8eac3ff018080ffff80ffff01ffff33ffa06b7a83babea1eec790c947db4464ab657dbe9b887fe9acc247062847b8c2a8a9ff830186a08080ff8080808080"
)

gen1 = b"\xff\x01" + original_generator
gen2 = b"\xff\x01\xff\x01" + original_generator
FAKE_BLOCK_HEIGHT1 = uint32(100)
FAKE_BLOCK_HEIGHT2 = uint32(200)

assert serialized_length(original_generator) == len(original_generator)
assert serialized_length(gen1) == len(gen1)
assert serialized_length(gen2) == len(gen2)


def spend_bundle_to_coin_spend_entry_list(bundle: SpendBundle) -> list[Any]:
    r = []
    for coin_spend in bundle.coin_spends:
        entry = [
            coin_spend.coin.parent_coin_info,
            sexp_from_stream(io.BytesIO(bytes(coin_spend.puzzle_reveal)), SExp.to),
            coin_spend.coin.amount,
            sexp_from_stream(io.BytesIO(bytes(coin_spend.solution)), SExp.to),
        ]
        r.append(entry)
    return r


class TestCompression:
    def test_compress_spend_bundle(self) -> None:
        pass


class TestDecompression:
    def test_deserialization(self) -> None:
        _cost, out = DESERIALIZE_MOD.run_with_cost(INFINITE_COST, [bytes(Program.to("hello"))])
        assert out == Program.to("hello")

    def test_deserialization_as_argument(self) -> None:
        _cost, out = TEST_GEN_DESERIALIZE.run_with_cost(
            INFINITE_COST, [DESERIALIZE_MOD, Nil, bytes(Program.to("hello"))]
        )
        print(bytes(Program.to("hello")))
        print()
        print(out)
        assert out == Program.to("hello")

    def test_decompress_puzzle(self) -> None:
        _cost, out = DECOMPRESS_PUZZLE.run_with_cost(
            INFINITE_COST, [DESERIALIZE_MOD, b"\xff", bytes(Program.to("pubkey")), b"\x80"]
        )

        print()
        print(out)

    # An empty CSE is invalid. (An empty CSE list may be okay)
    # def test_decompress_empty_cse(self):
    #    cse0 = binutils.assemble("()")
    #    cost, out = DECOMPRESS_CSE.run_with_cost(INFINITE_COST, [DESERIALIZE_MOD, DECOMPRESS_PUZZLE, b"\xff", b"\x80", cse0])
    #    print()
    #    print(out)

    def test_decompress_cse(self) -> None:
        """Decompress a single CSE / CoinSpendEntry"""
        cse0 = binutils.assemble(
            "((0x0000000000000000000000000000000000000000000000000000000000000000 0x0186a0) (0xb081963921826355dcb6c355ccf9c2637c18adf7d38ee44d803ea9ca41587e48c913d8d46896eb830aeadfc13144a8eac3 (() (q (51 0x6b7a83babea1eec790c947db4464ab657dbe9b887fe9acc247062847b8c2a8a9 0x0186a0)) ())))"
        )
        _cost, out = DECOMPRESS_CSE.run_with_cost(
            INFINITE_COST, [DESERIALIZE_MOD, DECOMPRESS_PUZZLE, b"\xff", b"\x80", cse0]
        )

        print()
        print(out)

    def test_decompress_cse_with_prefix(self) -> None:
        cse0 = binutils.assemble(
            "((0x0000000000000000000000000000000000000000000000000000000000000000 0x0186a0) (0xb081963921826355dcb6c355ccf9c2637c18adf7d38ee44d803ea9ca41587e48c913d8d46896eb830aeadfc13144a8eac3 (() (q (51 0x6b7a83babea1eec790c947db4464ab657dbe9b887fe9acc247062847b8c2a8a9 0x0186a0)) ())))"
        )

        start = 2 + 44
        end = start + 238
        prefix = original_generator[start:end]
        # (deserialize decompress_puzzle puzzle_prefix cse)
        _cost, out = DECOMPRESS_CSE_WITH_PREFIX.run_with_cost(
            INFINITE_COST, [DESERIALIZE_MOD, DECOMPRESS_PUZZLE, prefix, cse0]
        )

        print()
        print(out)

    def test_block_program_zero(self) -> None:
        "Decompress a list of CSEs"
        cse2 = binutils.assemble(
            """
(
  ((0x0000000000000000000000000000000000000000000000000000000000000000 0x0186a0)
   (0xb081963921826355dcb6c355ccf9c2637c18adf7d38ee44d803ea9ca41587e48c913d8d46896eb830aeadfc13144a8eac3
    (() (q (51 0x6b7a83babea1eec790c947db4464ab657dbe9b887fe9acc247062847b8c2a8a9 0x0186a0)) ()))
  )

  ((0x0000000000000000000000000000000000000000000000000000000000000001 0x0186a0)
   (0xb0a6207f5173ec41491d9f2c1b8fff5579e13703077e0eaca8fe587669dcccf51e9209a6b65576845ece5f7c2f3229e7e3
   (() (q (51 0x24254a3efc3ebfac9979bbe0d615e2eda043aa329905f65b63846fa24149e2b6 0x0186a0)) ())))

)
        """
        )

        start = 2 + 44
        end = start + 238

        # (mod (decompress_puzzle decompress_coin_spend_entry start end compressed_cses deserialize generator_list reserved_arg)
        # cost, out = DECOMPRESS_BLOCK.run_with_cost(INFINITE_COST, [DECOMPRESS_PUZZLE, DECOMPRESS_CSE, start, Program.to(end), cse0, DESERIALIZE_MOD, bytes(original_generator)])
        _cost, out = DECOMPRESS_BLOCK.run_with_cost(
            INFINITE_COST,
            [
                DECOMPRESS_PUZZLE,
                DECOMPRESS_CSE_WITH_PREFIX,
                start,
                Program.to(end),
                cse2,
                DESERIALIZE_MOD,
                [bytes(original_generator)],
            ],
        )

        print()
        print(out)

    def test_block_program_zero_with_curry(self) -> None:
        cse2 = binutils.assemble(
            """
(
  ((0x0000000000000000000000000000000000000000000000000000000000000000 0x0186a0)
   (0xb081963921826355dcb6c355ccf9c2637c18adf7d38ee44d803ea9ca41587e48c913d8d46896eb830aeadfc13144a8eac3
    (() (q (51 0x6b7a83babea1eec790c947db4464ab657dbe9b887fe9acc247062847b8c2a8a9 0x0186a0)) ()))
  )

  ((0x0000000000000000000000000000000000000000000000000000000000000001 0x0186a0)
   (0xb0a6207f5173ec41491d9f2c1b8fff5579e13703077e0eaca8fe587669dcccf51e9209a6b65576845ece5f7c2f3229e7e3
   (() (q (51 0x24254a3efc3ebfac9979bbe0d615e2eda043aa329905f65b63846fa24149e2b6 0x0186a0)) ())))

)
        """
        )

        start = 2 + 44
        end = start + 238

        # (mod (decompress_puzzle decompress_coin_spend_entry start end compressed_cses deserialize generator_list reserved_arg)
        # cost, out = DECOMPRESS_BLOCK.run_with_cost(INFINITE_COST, [DECOMPRESS_PUZZLE, DECOMPRESS_CSE, start, Program.to(end), cse0, DESERIALIZE_MOD, bytes(original_generator)])
        p = DECOMPRESS_BLOCK.curry(DECOMPRESS_PUZZLE, DECOMPRESS_CSE_WITH_PREFIX, start, Program.to(end))
        _cost, out = p.run_with_cost(INFINITE_COST, [cse2, DESERIALIZE_MOD, [bytes(original_generator)]])

        print()
        print(p)
        print(out)

        p_with_cses = DECOMPRESS_BLOCK.curry(
            DECOMPRESS_PUZZLE, DECOMPRESS_CSE_WITH_PREFIX, start, Program.to(end), cse2, DESERIALIZE_MOD
        )
        generator_args = Program.to([[original_generator]])
        _cost, out = p_with_cses.run_with_cost(INFINITE_COST, generator_args)

        print()
        print(p_with_cses)
        print(out)



from __future__ import annotations

from re import search
from typing import Union

from packaging.version import InvalidVersion, Version

from chik import __version__


def _chik_short_version_from_version(version: Version) -> str:
    release_version_str = f"{version.major}.{version.minor}.{version.micro}"

    return release_version_str if version.pre is None else release_version_str + "".join(map(str, version.pre))


def _chik_short_version_from_str(version: str) -> str:
    try:
        return chik_short_version(Version(version))
    except InvalidVersion:
        pass
    match = search(r"^(\d+\.\d+\.\d+)", version)
    if match is not None:
        return match.group(1)

    return version


def chik_short_version(version: Union[str, Version] = __version__) -> str:
    if isinstance(version, Version):
        return _chik_short_version_from_version(version)

    return _chik_short_version_from_str(version)



from __future__ import annotations

import logging
from pathlib import Path

import yaml

from chik.data_layer.data_layer_util import PluginRemote
from chik.util.log_exceptions import log_exceptions


async def load_plugin_configurations(root_path: Path, config_type: str, log: logging.Logger) -> list[PluginRemote]:
    """
    Loads plugin configurations from the specified directory and validates that the contents
    are in the expected JSON format (an array of PluginRemote objects). It gracefully handles errors
    and ensures that the necessary directories exist, creating them if they do not.

    Args:
        root_path (Path): The root path where the plugins directory is located.
        config_type (str): The type of plugins to load ('downloaders' or 'uploaders').

    Returns:
        list[PluginRemote]: A list of valid PluginRemote instances for the specified plugin type.
    """
    config_path = root_path / "plugins" / config_type
    config_path.mkdir(parents=True, exist_ok=True)  # Ensure the config directory exists

    valid_configs = []
    for conf_file in config_path.glob("*.conf"):
        with log_exceptions(
            log=log,
            consume=True,
            message=f"Skipping config file due to failure loading or parsing: {conf_file}",
        ):
            with open(conf_file) as file:
                data = yaml.safe_load(file)

            valid_configs.extend([PluginRemote.unmarshal(marshalled=item) for item in data])
            log.info(f"loaded plugin configuration: {conf_file}")
    return valid_configs



from __future__ import annotations

import asyncio
import sys
from io import TextIOWrapper
from typing import Optional

import click

from chik.cmds.cmd_classes import ChikCliContext
from chik.util.config import load_config


@click.group("passphrase", help="Manage your keyring passphrase")
def passphrase_cmd() -> None:
    pass


@passphrase_cmd.command(
    "set",
    help="""Sets or updates the keyring passphrase. If --passphrase-file and/or --current-passphrase-file options are
            provided, the passphrases will be read from the specified files. Otherwise, a prompt will be provided to
            enter the passphrase.""",
    short_help="Set or update the keyring passphrase",
)
@click.option("--passphrase-file", type=click.File("r"), help="File or descriptor to read the passphrase from")
@click.option(
    "--current-passphrase-file", type=click.File("r"), help="File or descriptor to read the current passphrase from"
)
@click.option("--hint", type=str, help="Passphrase hint")
@click.pass_context
def set_cmd(
    ctx: click.Context,
    passphrase_file: Optional[TextIOWrapper],
    current_passphrase_file: Optional[TextIOWrapper],
    hint: Optional[str],
) -> None:
    from chik.cmds.passphrase_funcs import (
        async_update_daemon_passphrase_cache_if_running,
        read_passphrase_from_file,
        set_or_update_passphrase,
        verify_passphrase_meets_requirements,
    )

    success: bool = False
    current_passphrase: Optional[str] = None
    if current_passphrase_file is not None:
        current_passphrase = read_passphrase_from_file(current_passphrase_file)

    if passphrase_file is not None:
        try:
            # Read the passphrase from a file and verify it
            new_passphrase: str = read_passphrase_from_file(passphrase_file)
            valid_passphrase, error_msg = verify_passphrase_meets_requirements(
                new_passphrase, new_passphrase
            )  # new_passphrase provided for both args since we don't have a separate confirmation passphrase

            if not valid_passphrase:
                raise ValueError(f"{error_msg}")
        except ValueError as e:
            print(f"Unable to set passphrase: {e}")
        except Exception as e:
            print(f"Failed to read passphrase: {e}")
        else:
            # Interactively prompt for the current passphrase (if set)
            success = set_or_update_passphrase(
                passphrase=new_passphrase, current_passphrase=current_passphrase, hint=hint
            )
    else:
        success = set_or_update_passphrase(passphrase=None, current_passphrase=current_passphrase, hint=hint)

    if success:
        # Attempt to update the daemon's passphrase cache
        root_path = ChikCliContext.set_default(ctx).root_path
        config = load_config(root_path, "config.yaml")
        sys.exit(asyncio.run(async_update_daemon_passphrase_cache_if_running(root_path, config)))


@passphrase_cmd.command(
    "remove",
    help="""Remove the keyring passphrase. If the --current-passphrase-file option is provided, the passphrase will be
            read from the specified file. Otherwise, a prompt will be provided to enter the passphrase.""",
    short_help="Remove the keyring passphrase",
)
@click.option(
    "--current-passphrase-file", type=click.File("r"), help="File or descriptor to read the current passphrase from"
)
@click.pass_context
def remove_cmd(ctx: click.Context, current_passphrase_file: Optional[TextIOWrapper]) -> None:
    from chik.cmds.passphrase_funcs import (
        async_update_daemon_passphrase_cache_if_running,
        read_passphrase_from_file,
        remove_passphrase,
    )

    current_passphrase: Optional[str] = None
    if current_passphrase_file is not None:
        current_passphrase = read_passphrase_from_file(current_passphrase_file)

    if remove_passphrase(current_passphrase):
        # Attempt to update the daemon's passphrase cache
        root_path = ChikCliContext.set_default(ctx).root_path
        config = load_config(root_path, "config.yaml")
        sys.exit(asyncio.run(async_update_daemon_passphrase_cache_if_running(root_path, config)))


@passphrase_cmd.group("hint", help="Manage the optional keyring passphrase hint")
def hint_cmd() -> None:
    pass


@hint_cmd.command("display", help="Display the keyring passphrase hint")
def display_hint() -> None:
    from chik.cmds.passphrase_funcs import display_passphrase_hint

    display_passphrase_hint()


@hint_cmd.command("set", help="Set or update the keyring passphrase hint")
@click.argument("hint", nargs=1)
def set_hint(hint: str) -> None:
    from chik.cmds.passphrase_funcs import set_passphrase_hint

    set_passphrase_hint(hint)


@hint_cmd.command("remove", help="Remove the keyring passphrase hint")
def remove_hint() -> None:
    from chik.cmds.passphrase_funcs import remove_passphrase_hint

    remove_passphrase_hint()



from __future__ import annotations

import asyncio
import asyncio.events
import asyncio.protocols
import dataclasses
import functools
import logging.config
import pathlib
import sys
import threading
from typing import Optional, final, overload

from chik._tests.util.misc import create_logger
from chik.server.chik_policy import ChikPolicy
from chik.server.start_service import async_run
from chik.util.task_referencer import create_referenced_task

if sys.platform == "win32":
    import _winapi

    NULL = _winapi.NULL


@final
@dataclasses.dataclass
class EchoServer(asyncio.Protocol):
    logger: logging.Logger

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        peername = transport.get_extra_info("peername")
        self.logger.info(f"connection from {peername}")
        self.transport = transport

    def data_received(self, data: bytes) -> None:
        print(f"data received: {data.hex()}")
        # TODO: review this
        self.transport.write(data)  # type: ignore[attr-defined]
        # print("and sent back")

        # close the socket
        # self.transport.close()


@overload
async def async_main(
    *,
    out_path: pathlib.Path,
    shutdown_path: pathlib.Path,
    ip: str = "127.0.0.1",
    port: int = 9678,
    port_holder: Optional[list[int]] = None,
) -> None: ...


@overload
async def async_main(
    *,
    out_path: pathlib.Path,
    thread_end_event: threading.Event,
    ip: str = "127.0.0.1",
    port: int = 9678,
    port_holder: Optional[list[int]] = None,
) -> None: ...


async def async_main(
    *,
    out_path: pathlib.Path,
    shutdown_path: Optional[pathlib.Path] = None,
    thread_end_event: Optional[threading.Event] = None,
    ip: str = "127.0.0.1",
    port: int = 9678,
    port_holder: Optional[list[int]] = None,
) -> None:
    with out_path.open(mode="w") as file:
        logger = create_logger(file=file)
        file_task: Optional[asyncio.Task[None]] = None
        if thread_end_event is None:
            assert shutdown_path is not None
            thread_end_event = threading.Event()

            async def dun() -> None:
                # TODO: switch to event driven code
                while shutdown_path.exists():  # noqa: ASYNC110
                    await asyncio.sleep(0.25)

                thread_end_event.set()

            file_task = create_referenced_task(dun())

        loop = asyncio.get_event_loop()
        server = await loop.create_server(functools.partial(EchoServer, logger=logger), ip, port)
        if port_holder is not None:
            [server_socket] = server.sockets
            # TODO: review if this is general enough, such as for ipv6
            port_holder.append(server_socket.getsockname()[1])
        logger.info(f"serving on {server.sockets[0].getsockname()}")

        try:
            try:
                # TODO: switch to event driven code
                while not thread_end_event.is_set():  # noqa: ASYNC110
                    await asyncio.sleep(0.1)
            finally:
                # the test checks explicitly for this
                logger.info("exit: shutting down")
            logger.info("exit: thread end event set")
        except KeyboardInterrupt:
            logger.info("exit: keyboard interrupt")
        except asyncio.CancelledError:
            logger.info("exit: cancelled")
        finally:
            logger.info("closing server")
            server.close()
            await server.wait_closed()
            logger.info("server closed")
            if file_task is not None:
                await file_task


def main(connection_limit: int = 25) -> None:
    asyncio.set_event_loop_policy(ChikPolicy())
    shutdown_path = pathlib.Path(sys.argv[1])
    async_run(
        async_main(
            shutdown_path=shutdown_path,
            out_path=shutdown_path.with_suffix(".out"),
        ),
        connection_limit=connection_limit - 100,
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Collection
from typing import Optional

from chikbip158 import PyBIP158

from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.full_block import FullBlock
from chik.types.header_block import HeaderBlock
from chik.types.spend_bundle_conditions import SpendBundleConditions
from chik.util.ints import uint64


def get_block_header(
    block: FullBlock, removals_and_additions: Optional[tuple[Collection[bytes32], Collection[Coin]]] = None
) -> HeaderBlock:
    """
    Returns a HeaderBlock from a FullBlock.
    If `removals_and_additions` is not None, account for them, as well as
    reward coins, in the creation of the transactions filter, otherwise create
    an empty one.
    """
    # Guard against passing removals and additions for a non-transaction block.
    assert removals_and_additions is None or block.is_transaction_block()
    # Create an empty filter to begin with
    byte_array_tx: list[bytearray] = []
    if removals_and_additions is not None and block.is_transaction_block():
        removals_names, tx_addition_coins = removals_and_additions
        for coin in tx_addition_coins:
            byte_array_tx.append(bytearray(coin.puzzle_hash))
        for coin in block.get_included_reward_coins():
            byte_array_tx.append(bytearray(coin.puzzle_hash))
        for name in removals_names:
            byte_array_tx.append(bytearray(name))

    bip158: PyBIP158 = PyBIP158(byte_array_tx)
    encoded_filter: bytes = bytes(bip158.GetEncoded())

    return HeaderBlock(
        block.finished_sub_slots,
        block.reward_chain_block,
        block.challenge_chain_sp_proof,
        block.challenge_chain_ip_proof,
        block.reward_chain_sp_proof,
        block.reward_chain_ip_proof,
        block.infused_challenge_chain_ip_proof,
        block.foliage,
        block.foliage_transaction_block,
        encoded_filter,
        block.transactions_info,
    )


def tx_removals_and_additions(results: Optional[SpendBundleConditions]) -> tuple[list[bytes32], list[Coin]]:
    """
    Doesn't return farmer and pool reward.
    """

    removals: list[bytes32] = []
    additions: list[Coin] = []

    # build removals list
    if results is None:
        return [], []
    for spend in results.spends:
        removals.append(bytes32(spend.coin_id))
        for puzzle_hash, amount, _ in spend.create_coin:
            additions.append(Coin(bytes32(spend.coin_id), bytes32(puzzle_hash), uint64(amount)))

    return removals, additions



from __future__ import annotations

import pytest

from chik.server.server import ChikServer
from chik.simulator.block_tools import BlockTools
from chik.simulator.full_node_simulator import FullNodeSimulator
from chik.types.peer_info import PeerInfo
from chik.wallet.puzzles.clawback.puzzle_decorator import ClawbackPuzzleDecorator
from chik.wallet.util.puzzle_decorator import PuzzleDecoratorManager
from chik.wallet.wallet_node import WalletNode


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_missing_decorator(
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
    trusted: bool,
    self_hostname: str,
) -> None:
    full_nodes, wallets, _ = simulator_and_wallet
    full_node_api = full_nodes[0]
    server_1: ChikServer = full_node_api.full_node.server
    wallet_node, server_2 = wallets[0]
    decorator_config = {"clawback_timelock": 3600}
    wallet_node.wallet_state_manager.decorator_manager = PuzzleDecoratorManager.create([decorator_config])
    await server_2.start_client(PeerInfo(self_hostname, server_1.get_port()), None)
    assert len(wallet_node.wallet_state_manager.decorator_manager.decorator_list) == 0


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_unknown_decorator(
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
    trusted: bool,
    self_hostname: str,
) -> None:
    full_nodes, wallets, _ = simulator_and_wallet
    full_node_api = full_nodes[0]
    server_1: ChikServer = full_node_api.full_node.server
    wallet_node, server_2 = wallets[0]
    decorator_config = {"decorator": "UNKNOWN", "clawback_timelock": 3600}
    wallet_node.wallet_state_manager.decorator_manager = PuzzleDecoratorManager.create([decorator_config])
    await server_2.start_client(PeerInfo(self_hostname, server_1.get_port()), None)
    assert len(wallet_node.wallet_state_manager.decorator_manager.decorator_list) == 0


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_decorator(
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
    trusted: bool,
    self_hostname: str,
) -> None:
    full_nodes, wallets, _ = simulator_and_wallet
    full_node_api = full_nodes[0]
    server_1: ChikServer = full_node_api.full_node.server
    wallet_node, server_2 = wallets[0]
    wallet = wallet_node.wallet_state_manager.main_wallet
    print(wallet_node.logged_in_fingerprint)
    decorator_config = {"decorator": "CLAWBACK", "clawback_timelock": 3600}
    wallet_node.wallet_state_manager.decorator_manager = PuzzleDecoratorManager.create([decorator_config])
    await server_2.start_client(PeerInfo(self_hostname, server_1.get_port()), None)
    assert len(wallet_node.wallet_state_manager.decorator_manager.decorator_list) == 1
    assert isinstance(wallet_node.wallet_state_manager.decorator_manager.decorator_list[0], ClawbackPuzzleDecorator)
    clawback_decorator: ClawbackPuzzleDecorator = wallet_node.wallet_state_manager.decorator_manager.decorator_list[0]
    assert clawback_decorator.time_lock == 3600
    puzzle = await wallet.get_new_puzzle()
    assert puzzle == wallet_node.wallet_state_manager.decorator_manager.decorate(puzzle)



from __future__ import annotations

from typing import Optional, Union

from chik.consensus.constants import ConsensusConstants
from chik.consensus.pot_iterations import calculate_ip_iters, calculate_iterations_quality, calculate_sp_iters
from chik.types.blockchain_format.proof_of_space import verify_and_get_quality_string
from chik.types.blockchain_format.reward_chain_block import RewardChainBlock, RewardChainBlockUnfinished
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint32, uint64


def iters_from_block(
    constants: ConsensusConstants,
    reward_chain_block: Union[RewardChainBlock, RewardChainBlockUnfinished],
    sub_slot_iters: uint64,
    difficulty: uint64,
    height: uint32,
) -> tuple[uint64, uint64]:
    if reward_chain_block.challenge_chain_sp_vdf is None:
        assert reward_chain_block.signage_point_index == 0
        cc_sp: bytes32 = reward_chain_block.pos_ss_cc_challenge_hash
    else:
        cc_sp = reward_chain_block.challenge_chain_sp_vdf.output.get_hash()

    quality_string: Optional[bytes32] = verify_and_get_quality_string(
        reward_chain_block.proof_of_space,
        constants,
        reward_chain_block.pos_ss_cc_challenge_hash,
        cc_sp,
        height=height,
    )
    assert quality_string is not None

    required_iters: uint64 = calculate_iterations_quality(
        constants.DIFFICULTY_CONSTANT_FACTOR,
        quality_string,
        reward_chain_block.proof_of_space.size,
        difficulty,
        cc_sp,
    )
    return (
        calculate_sp_iters(constants, sub_slot_iters, reward_chain_block.signage_point_index),
        calculate_ip_iters(
            constants,
            sub_slot_iters,
            reward_chain_block.signage_point_index,
            required_iters,
        ),
    )



from __future__ import annotations

import sys
from pathlib import Path
from typing import Any, Optional

import pytest
from _pytest.capture import CaptureFixture
from click.testing import CliRunner
from pytest_mock import MockerFixture

from chik.cmds.chik import cli
from chik.cmds.start_funcs import create_start_daemon_connection, launch_start_daemon


@pytest.mark.anyio
@pytest.mark.parametrize("skip_keyring", [False, True])
@pytest.mark.parametrize("unlock_keyring", [False, True])
async def test_daemon(
    skip_keyring: bool, unlock_keyring: bool, mocker: MockerFixture, capsys: CaptureFixture[str]
) -> None:
    class DummyConnection:
        @staticmethod
        async def is_keyring_locked() -> bool:
            return unlock_keyring

        @staticmethod
        async def unlock_keyring(_passphrase: str) -> bool:
            return True

    async def connect_to_daemon_and_validate(_root_path: Path, _config: dict[str, Any]) -> DummyConnection:
        return DummyConnection()

    class DummyKeychain:
        @staticmethod
        def get_cached_master_passphrase() -> Optional[str]:
            return None

    def get_current_passphrase() -> Optional[str]:
        return "a-passphrase"

    mocker.patch("chik.cmds.start_funcs.connect_to_daemon_and_validate", side_effect=connect_to_daemon_and_validate)
    mocker.patch("chik.cmds.start_funcs.Keychain", new=DummyKeychain)
    mocker.patch("chik.cmds.start_funcs.get_current_passphrase", side_effect=get_current_passphrase)

    daemon = await create_start_daemon_connection(Path("/path-not-exist"), {}, skip_keyring=skip_keyring)
    assert daemon is not None
    captured = capsys.readouterr()
    assert captured.err == ""
    if skip_keyring:
        assert captured.out.endswith("Skipping to unlock keyring\n")
    else:
        assert not captured.out.endswith("Skipping to unlock keyring\n")


@pytest.mark.anyio
def test_launch_start_daemon(tmp_path: Path) -> None:
    sys.argv[0] = "not-exist"
    with pytest.raises(FileNotFoundError):
        launch_start_daemon(tmp_path)

    helper: Path = Path(sys.executable)
    sys.argv[0] = str(helper.parent) + "/chik"
    process = launch_start_daemon(tmp_path)
    assert process is not None
    process.kill()
    process.wait()


def test_start_daemon(tmp_path: Path, empty_keyring: Any, mocker: MockerFixture) -> None:
    class DummyDaemon:
        @staticmethod
        async def close() -> None:
            return None

    async def create_start_daemon_connection_dummy(
        root_path: Path, config: dict[str, Any], *, skip_keyring: bool
    ) -> DummyDaemon:
        return DummyDaemon()

    mocker.patch(
        "chik.cmds.start_funcs.create_start_daemon_connection", side_effect=create_start_daemon_connection_dummy
    )

    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["--root-path", str(tmp_path), "init"],
    )
    assert result.exit_code == 0
    result = runner.invoke(cli, ["--root-path", str(tmp_path), "start", "daemon", "-s"])
    assert result.exit_code == 0



from __future__ import annotations

from collections.abc import Iterable

from chik.types.blockchain_format.sized_bytes import bytes32


class IntegrityError(Exception):
    pass


def build_message_with_hashes(message: str, bytes_objects: Iterable[bytes]) -> str:
    return "\n".join([message, *[f"    {b.hex()}" for b in bytes_objects]])


class TreeGenerationIncrementingError(IntegrityError):
    def __init__(self, store_ids: list[bytes32]) -> None:
        super().__init__(
            build_message_with_hashes(
                message="Found trees with generations not properly incrementing:",
                bytes_objects=store_ids,
            )
        )


class NodeHashError(IntegrityError):
    def __init__(self, node_hashes: list[bytes32]) -> None:
        super().__init__(
            build_message_with_hashes(
                message="Found nodes with incorrect hashes:",
                bytes_objects=node_hashes,
            )
        )


class KeyNotFoundError(Exception):
    def __init__(self, key: bytes) -> None:
        super().__init__(f"Key not found: {key.hex()}")


class OfferIntegrityError(Exception):
    pass


class ProofIntegrityError(Exception):
    pass


class LauncherCoinNotFoundError(Exception):
    pass



from __future__ import annotations

import random
from dataclasses import dataclass
from typing import Optional

import pytest
from chik_rs import G1Element

from chik._tests.util.misc import Marks, datacases
from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.types.blockchain_format.proof_of_space import ProofOfSpace, passes_plot_filter, verify_and_get_quality_string
from chik.types.blockchain_format.sized_bytes import bytes32, bytes48
from chik.util.ints import uint8, uint32


@dataclass
class ProofOfSpaceCase:
    id: str
    pos_challenge: bytes32
    plot_size: uint8
    plot_public_key: G1Element
    pool_public_key: Optional[G1Element] = None
    pool_contract_puzzle_hash: Optional[bytes32] = None
    height: uint32 = uint32(0)
    expected_error: Optional[str] = None
    marks: Marks = ()


@datacases(
    ProofOfSpaceCase(
        id="Neither pool public key nor pool contract puzzle hash",
        pos_challenge=bytes32(b"1" * 32),
        plot_size=uint8(0),
        plot_public_key=G1Element(),
        expected_error="Expected pool public key or pool contract puzzle hash but got neither",
    ),
    ProofOfSpaceCase(
        id="Both pool public key and pool contract puzzle hash",
        pos_challenge=bytes32(b"1" * 32),
        plot_size=uint8(0),
        plot_public_key=G1Element(),
        pool_public_key=G1Element(),
        pool_contract_puzzle_hash=bytes32(b"1" * 32),
        expected_error="Expected pool public key or pool contract puzzle hash but got both",
    ),
    ProofOfSpaceCase(
        id="Lower than minimum plot size",
        pos_challenge=bytes32(b"1" * 32),
        plot_size=uint8(31),
        plot_public_key=G1Element(),
        pool_public_key=G1Element(),
        expected_error="Plot size is lower than the minimum",
    ),
    ProofOfSpaceCase(
        id="Higher than maximum plot size",
        pos_challenge=bytes32(b"1" * 32),
        plot_size=uint8(51),
        plot_public_key=G1Element(),
        pool_public_key=G1Element(),
        expected_error="Plot size is higher than the maximum",
    ),
    ProofOfSpaceCase(
        id="Different challenge",
        pos_challenge=bytes32(b"1" * 32),
        plot_size=uint8(42),
        pool_public_key=G1Element(),
        plot_public_key=G1Element(),
        expected_error="Calculated pos challenge doesn't match the provided one",
    ),
    ProofOfSpaceCase(
        id="Not passing the plot filter with size 9",
        pos_challenge=bytes32.from_hexstr("08b23cc2844dfb92d2eedaa705a1ce665d571ee753bd81cbb67b92caa6d34722"),
        plot_size=uint8(42),
        pool_public_key=G1Element.from_bytes_unchecked(
            bytes48.from_hexstr(
                "b6449c2c68df97c19e884427e42ee7350982d4020571ead08732615ff39bd216bfd630b6460784982bec98b49fea79d0"
            )
        ),
        plot_public_key=G1Element.from_bytes_unchecked(
            bytes48.from_hexstr(
                "b17d368f5400230b2b01464807825bf4163c5c159bd7d4465f935912e538ac9fb996dd9a9c479bd8aa6256bdca1fed96"
            )
        ),
        height=uint32(5495999),
        expected_error="Did not pass the plot filter",
    ),
    ProofOfSpaceCase(
        id="Passing the plot filter with size 8",
        pos_challenge=bytes32.from_hexstr("08b23cc2844dfb92d2eedaa705a1ce665d571ee753bd81cbb67b92caa6d34722"),
        plot_size=uint8(42),
        pool_public_key=G1Element.from_bytes_unchecked(
            bytes48.from_hexstr(
                "b6449c2c68df97c19e884427e42ee7350982d4020571ead08732615ff39bd216bfd630b6460784982bec98b49fea79d0"
            )
        ),
        plot_public_key=G1Element.from_bytes_unchecked(
            bytes48.from_hexstr(
                "b17d368f5400230b2b01464807825bf4163c5c159bd7d4465f935912e538ac9fb996dd9a9c479bd8aa6256bdca1fed96"
            )
        ),
        height=uint32(5496000),
    ),
)
def test_verify_and_get_quality_string(caplog: pytest.LogCaptureFixture, case: ProofOfSpaceCase) -> None:
    pos = ProofOfSpace(
        challenge=case.pos_challenge,
        pool_public_key=case.pool_public_key,
        pool_contract_puzzle_hash=case.pool_contract_puzzle_hash,
        plot_public_key=case.plot_public_key,
        size=case.plot_size,
        proof=b"1",
    )
    quality_string = verify_and_get_quality_string(
        pos=pos,
        constants=DEFAULT_CONSTANTS,
        original_challenge_hash=bytes32.from_hexstr(
            "0x73490e166d0b88347c37d921660b216c27316aae9a3450933d3ff3b854e5831a"
        ),
        signage_point=bytes32.from_hexstr("0x7b3e23dbd438f9aceefa9827e2c5538898189987f49b06eceb7a43067e77b531"),
        height=case.height,
    )
    assert quality_string is None
    assert len(caplog.text) == 0 if case.expected_error is None else case.expected_error in caplog.text


class TestProofOfSpace:
    @pytest.mark.parametrize("prefix_bits", [DEFAULT_CONSTANTS.NUMBER_ZERO_BITS_PLOT_FILTER, 8, 7, 6, 5, 1, 0])
    def test_can_create_proof(self, prefix_bits: int, seeded_random: random.Random) -> None:
        """
        Tests that the change of getting a correct proof is exactly 1/target_filter.
        """
        num_trials = 100000
        success_count = 0
        target_filter = 2**prefix_bits
        for _ in range(num_trials):
            challenge_hash = bytes32.random(seeded_random)
            plot_id = bytes32.random(seeded_random)
            sp_output = bytes32.random(seeded_random)

            if passes_plot_filter(prefix_bits, plot_id, challenge_hash, sp_output):
                success_count += 1

        assert abs((success_count * target_filter / num_trials) - 1) < 0.35



from __future__ import annotations

import enum


class SortKey(enum.Enum):
    CONFIRMED_AT_HEIGHT = "ORDER BY confirmed_at_height {ASC}"
    RELEVANCE = "ORDER BY confirmed {ASC}, confirmed_at_height {DESC}, created_at_time {DESC}"

    def ascending(self) -> str:
        return self.value.format(ASC="ASC", DESC="DESC")

    def descending(self) -> str:
        return self.value.format(ASC="DESC", DESC="ASC")



from __future__ import annotations

import dataclasses
from typing import Optional

import click
import pytest
from chik_rs import AugSchemeMPL, G1Element, G2Element, PrivateKey
from click.testing import CliRunner

from chik._tests.cmds.test_cmd_framework import check_click_parsing
from chik._tests.cmds.wallet.test_consts import STD_TX
from chik._tests.environments.wallet import WalletStateTransition, WalletTestFramework
from chik.cmds.cmd_classes import chik_command
from chik.cmds.cmd_helpers import NeedsWalletRPC, TransactionsIn, TransactionsOut, WalletClientInfo
from chik.cmds.cmds_util import TransactionBundle
from chik.cmds.signer import (
    ApplySignaturesCMD,
    ExecuteSigningInstructionsCMD,
    GatherSigningInfoCMD,
    PushTransactionsCMD,
    QrCodeDisplay,
    SPIn,
    SPOut,
)
from chik.rpc.util import ALL_TRANSLATION_LAYERS
from chik.rpc.wallet_request_types import (
    ApplySignatures,
    ExecuteSigningInstructions,
    GatherSigningInfo,
    GatherSigningInfoResponse,
    SubmitTransactions,
)
from chik.rpc.wallet_rpc_client import WalletRpcClient
from chik.types.blockchain_format.coin import Coin as ConsensusCoin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend, make_spend
from chik.util.hash import std_hash
from chik.util.ints import uint64
from chik.util.streamable import Streamable
from chik.wallet.conditions import AggSigMe
from chik.wallet.derivation_record import DerivationRecord
from chik.wallet.derive_keys import _derive_path_unhardened
from chik.wallet.puzzles.p2_delegated_puzzle_or_hidden_puzzle import (
    DEFAULT_HIDDEN_PUZZLE_HASH,
    calculate_synthetic_offset,
)
from chik.wallet.signer_protocol import (
    Coin,
    KeyHints,
    PathHint,
    SignedTransaction,
    SigningInstructions,
    SigningResponse,
    SigningTarget,
    Spend,
    SumHint,
    TransactionInfo,
    UnsignedTransaction,
)
from chik.wallet.util.blind_signer_tl import (
    BLIND_SIGNER_TRANSLATION,
    BSTLPathHint,
    BSTLSigningInstructions,
    BSTLSigningResponse,
    BSTLSigningTarget,
    BSTLSumHint,
    BSTLUnsignedTransaction,
)
from chik.wallet.util.klvm_streamable import (
    TranslationLayer,
    TranslationLayerMapping,
    byte_serialize_klvm_streamable,
    klvm_streamable,
    json_deserialize_with_klvm_streamable,
    json_serialize_with_klvm_streamable,
)
from chik.wallet.util.tx_config import DEFAULT_TX_CONFIG
from chik.wallet.wallet import Wallet
from chik.wallet.wallet_spend_bundle import WalletSpendBundle
from chik.wallet.wallet_state_manager import WalletStateManager


def test_unsigned_transaction_type() -> None:
    pubkey: G1Element = G1Element()
    message: bytes = b"message"

    coin: ConsensusCoin = ConsensusCoin(bytes32.zeros, bytes32.zeros, uint64(0))
    puzzle: Program = Program.to(1)
    solution: Program = Program.to([AggSigMe(pubkey, message).to_program()])

    coin_spend: CoinSpend = make_spend(coin, puzzle, solution)
    assert Spend.from_coin_spend(coin_spend).as_coin_spend() == coin_spend

    tx: UnsignedTransaction = UnsignedTransaction(
        TransactionInfo([Spend.from_coin_spend(coin_spend)]),
        SigningInstructions(
            KeyHints([], []),
            [SigningTarget(pubkey.get_fingerprint().to_bytes(4, "big"), message, bytes32([1] * 32))],
        ),
    )

    assert tx == json_deserialize_with_klvm_streamable(json_serialize_with_klvm_streamable(tx), UnsignedTransaction)
    as_json_dict = {
        "coin": {
            "parent_coin_id": "0x" + tx.transaction_info.spends[0].coin.parent_coin_id.hex(),
            "puzzle_hash": "0x" + tx.transaction_info.spends[0].coin.puzzle_hash.hex(),
            "amount": tx.transaction_info.spends[0].coin.amount,
        },
        "puzzle": "0x" + bytes(tx.transaction_info.spends[0].puzzle).hex(),
        "solution": "0x" + bytes(tx.transaction_info.spends[0].solution).hex(),
    }
    assert tx.transaction_info.spends[0].to_json_dict() == as_json_dict


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [1],
            "trusted": True,
            "reuse_puzhash": True,
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_p2dohp_wallet_signer_protocol(wallet_environments: WalletTestFramework) -> None:
    wallet: Wallet = wallet_environments.environments[0].xck_wallet
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client

    # Test first that we can properly examine and sign a regular transaction
    async with wallet.wallet_state_manager.new_action_scope(wallet_environments.tx_config, push=False) as action_scope:
        [coin] = await wallet.select_coins(uint64(0), action_scope)
    puzzle: Program = await wallet.puzzle_for_puzzle_hash(coin.puzzle_hash)
    delegated_puzzle: Program = Program.to(None)
    delegated_puzzle_hash: bytes32 = delegated_puzzle.get_tree_hash()
    solution: Program = Program.to([None, None, None])

    coin_spend: CoinSpend = make_spend(
        coin,
        puzzle,
        solution,
    )

    derivation_record: Optional[
        DerivationRecord
    ] = await wallet_state_manager.puzzle_store.get_derivation_record_for_puzzle_hash(coin.puzzle_hash)
    assert derivation_record is not None
    pubkey: G1Element = derivation_record.pubkey
    atom = puzzle.uncurry()[1].at("f").atom
    assert atom is not None
    synthetic_pubkey: G1Element = G1Element.from_bytes(atom)
    message: bytes = delegated_puzzle_hash + coin.name() + wallet_state_manager.constants.AGG_SIG_ME_ADDITIONAL_DATA

    utx: UnsignedTransaction = UnsignedTransaction(
        TransactionInfo([Spend.from_coin_spend(coin_spend)]),
        (
            await wallet_rpc.gather_signing_info(GatherSigningInfo(spends=[Spend.from_coin_spend(coin_spend)]))
        ).signing_instructions,
    )
    assert utx.signing_instructions.key_hints.sum_hints == [
        SumHint(
            [pubkey.get_fingerprint().to_bytes(4, "big")],
            calculate_synthetic_offset(pubkey, DEFAULT_HIDDEN_PUZZLE_HASH).to_bytes(32, "big"),
            wallet_state_manager.main_wallet.puzzle_for_pk(pubkey).uncurry()[1].at("f").as_atom(),
        )
    ]
    assert utx.signing_instructions.key_hints.path_hints == [
        PathHint(
            wallet_state_manager.root_pubkey.get_fingerprint().to_bytes(4, "big"),
            [uint64(12381), uint64(8444), uint64(2), uint64(derivation_record.index)],
        )
    ]
    assert len(utx.signing_instructions.targets) == 1
    assert utx.signing_instructions.targets[0].fingerprint == synthetic_pubkey.get_fingerprint().to_bytes(4, "big")
    assert utx.signing_instructions.targets[0].message == message

    signing_responses: list[SigningResponse] = (
        await wallet_rpc.execute_signing_instructions(ExecuteSigningInstructions(utx.signing_instructions))
    ).signing_responses
    assert len(signing_responses) == 1
    assert signing_responses[0].hook == utx.signing_instructions.targets[0].hook
    assert AugSchemeMPL.verify(synthetic_pubkey, message, G2Element.from_bytes(signing_responses[0].signature))

    # Now test that we can partially sign a transaction
    ACS: Program = Program.to(1)
    ACS_PH = Program.to(1).get_tree_hash()
    not_our_private_key: PrivateKey = PrivateKey.from_bytes(bytes([1] * 32))
    not_our_pubkey: G1Element = not_our_private_key.get_g1()
    not_our_message: bytes = b"not our message"
    not_our_coin: ConsensusCoin = ConsensusCoin(
        bytes32.zeros,
        ACS_PH,
        uint64(0),
    )
    not_our_coin_spend: CoinSpend = make_spend(not_our_coin, ACS, Program.to([[49, not_our_pubkey, not_our_message]]))

    not_our_utx: UnsignedTransaction = UnsignedTransaction(
        TransactionInfo([Spend.from_coin_spend(coin_spend), Spend.from_coin_spend(not_our_coin_spend)]),
        (
            await wallet_rpc.gather_signing_info(
                GatherSigningInfo(spends=[Spend.from_coin_spend(coin_spend), Spend.from_coin_spend(not_our_coin_spend)])
            )
        ).signing_instructions,
    )
    assert not_our_utx.signing_instructions.key_hints == utx.signing_instructions.key_hints
    assert len(not_our_utx.signing_instructions.targets) == 2
    assert not_our_utx.signing_instructions.targets[0].fingerprint == synthetic_pubkey.get_fingerprint().to_bytes(
        4, "big"
    )
    assert not_our_utx.signing_instructions.targets[0].message == bytes(message)
    assert not_our_utx.signing_instructions.targets[1].fingerprint == not_our_pubkey.get_fingerprint().to_bytes(
        4, "big"
    )
    assert not_our_utx.signing_instructions.targets[1].message == bytes(not_our_message)
    not_our_signing_instructions: SigningInstructions = not_our_utx.signing_instructions
    with pytest.raises(ValueError, match=r"not found \(or path/sum hinted to\)"):
        await wallet_state_manager.execute_signing_instructions(not_our_signing_instructions)
    with pytest.raises(ValueError, match=r"No pubkey found \(or path hinted to\) for fingerprint"):
        await wallet_state_manager.execute_signing_instructions(
            dataclasses.replace(
                not_our_signing_instructions,
                key_hints=dataclasses.replace(
                    not_our_signing_instructions.key_hints,
                    sum_hints=[
                        *not_our_signing_instructions.key_hints.sum_hints,
                        SumHint([bytes(not_our_pubkey)], std_hash(b"sum hint only"), bytes(G1Element())),
                    ],
                ),
            )
        )
    with pytest.raises(ValueError, match="No root pubkey for fingerprint"):
        await wallet_state_manager.execute_signing_instructions(
            dataclasses.replace(
                not_our_signing_instructions,
                key_hints=dataclasses.replace(
                    not_our_signing_instructions.key_hints,
                    path_hints=[
                        *not_our_signing_instructions.key_hints.path_hints,
                        PathHint(bytes(not_our_pubkey), [uint64(0)]),
                    ],
                ),
            )
        )
    signing_responses_2 = await wallet_state_manager.execute_signing_instructions(
        not_our_signing_instructions, partial_allowed=True
    )
    assert len(signing_responses_2) == 2
    assert (
        bytes(AugSchemeMPL.aggregate([G2Element.from_bytes(sig.signature) for sig in signing_responses_2]))
        == signing_responses[0].signature
    )

    signed_txs: list[SignedTransaction] = (
        await wallet_rpc.apply_signatures(
            ApplySignatures(spends=[Spend.from_coin_spend(coin_spend)], signing_responses=signing_responses)
        )
    ).signed_transactions
    await wallet_rpc.submit_transactions(SubmitTransactions(signed_transactions=signed_txs))
    await wallet_environments.full_node.wait_bundle_ids_in_mempool(
        [
            WalletSpendBundle(
                [spend.as_coin_spend() for tx in signed_txs for spend in tx.transaction_info.spends],
                G2Element.from_bytes(signing_responses[0].signature),
            ).name()
        ]
    )

    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                # We haven't submitted a TransactionRecord so the wallet won't know about this until confirmed
                pre_block_balance_updates={},
                post_block_balance_updates={
                    1: {
                        "confirmed_wallet_balance": -1 * coin.amount,
                        "unconfirmed_wallet_balance": -1 * coin.amount,
                        "spendable_balance": -1 * coin.amount,
                        "max_send_amount": -1 * coin.amount,
                        "unspent_coin_count": -1,
                    },
                },
            ),
        ]
    )

    # And test that we can get compressed versions if we want
    request = GatherSigningInfo(
        [Spend.from_coin_spend(coin_spend), Spend.from_coin_spend(not_our_coin_spend)]
    ).to_json_dict()
    response_dict = await wallet_rpc.fetch("gather_signing_info", {"translation": "chip-0029", **request})
    response: GatherSigningInfoResponse = json_deserialize_with_klvm_streamable(
        response_dict, GatherSigningInfoResponse, translation_layer=BLIND_SIGNER_TRANSLATION
    )
    assert response.signing_instructions == not_our_utx.signing_instructions


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [1],
            "trusted": True,
            "reuse_puzhash": True,
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_p2blsdohp_execute_signing_instructions(wallet_environments: WalletTestFramework) -> None:
    wallet: Wallet = wallet_environments.environments[0].xck_wallet
    root_sk: PrivateKey = wallet.wallet_state_manager.get_master_private_key()
    root_pk: G1Element = root_sk.get_g1()
    root_fingerprint: bytes = root_pk.get_fingerprint().to_bytes(4, "big")

    # Test just a path hint
    test_name: bytes32 = std_hash(b"path hint only")
    child_sk: PrivateKey = _derive_path_unhardened(root_sk, [uint64(1), uint64(2), uint64(3), uint64(4)])
    signing_responses: list[SigningResponse] = await wallet.execute_signing_instructions(
        SigningInstructions(
            KeyHints(
                [],
                [PathHint(root_fingerprint, [uint64(1), uint64(2), uint64(3), uint64(4)])],
            ),
            [SigningTarget(child_sk.get_g1().get_fingerprint().to_bytes(4, "big"), test_name, test_name)],
        )
    )
    assert signing_responses == [SigningResponse(bytes(AugSchemeMPL.sign(child_sk, test_name)), test_name)]

    # Test just a sum hint
    test_name = std_hash(b"sum hint only")
    other_sk: PrivateKey = PrivateKey.from_bytes(test_name)
    sum_pk: G1Element = other_sk.get_g1() + root_pk
    signing_instructions: SigningInstructions = SigningInstructions(
        KeyHints(
            [SumHint([root_fingerprint], test_name, bytes(sum_pk))],
            [],
        ),
        [SigningTarget(sum_pk.get_fingerprint().to_bytes(4, "big"), test_name, test_name)],
    )
    for partial_allowed in (True, False):
        signing_responses = await wallet.execute_signing_instructions(signing_instructions, partial_allowed)
        assert signing_responses == [
            SigningResponse(
                bytes(
                    AugSchemeMPL.aggregate(
                        [
                            AugSchemeMPL.sign(other_sk, test_name, sum_pk),
                            AugSchemeMPL.sign(root_sk, test_name, sum_pk),
                        ]
                    )
                ),
                test_name,
            ),
        ]
    # Toss in a random SigningTarget to see that the responses split up
    signing_instructions = dataclasses.replace(
        signing_instructions,
        targets=[
            SigningTarget(sum_pk.get_fingerprint().to_bytes(4, "big"), test_name, test_name),
            SigningTarget(b"random fingerprint", test_name, test_name),
        ],
    )
    signing_responses = await wallet.execute_signing_instructions(signing_instructions, partial_allowed=True)
    assert signing_responses == [
        SigningResponse(
            bytes(
                AugSchemeMPL.sign(root_sk, test_name, sum_pk),
            ),
            test_name,
        ),
        SigningResponse(
            bytes(
                AugSchemeMPL.sign(other_sk, test_name, sum_pk),
            ),
            test_name,
        ),
    ]

    # Test both path and sum hint
    test_name = std_hash(b"path and sum hint")
    child_sk = _derive_path_unhardened(root_sk, [uint64(1), uint64(2), uint64(3), uint64(4)])
    other_sk = PrivateKey.from_bytes(test_name)
    sum_pk = child_sk.get_g1() + other_sk.get_g1()
    signing_instructions = SigningInstructions(
        KeyHints(
            [SumHint([child_sk.get_g1().get_fingerprint().to_bytes(4, "big")], test_name, bytes(sum_pk))],
            [PathHint(root_fingerprint, [uint64(1), uint64(2), uint64(3), uint64(4)])],
        ),
        [SigningTarget(sum_pk.get_fingerprint().to_bytes(4, "big"), test_name, test_name)],
    )
    for partial_allowed in (True, False):
        signing_responses = await wallet.execute_signing_instructions(signing_instructions, partial_allowed)
        assert signing_responses == [
            SigningResponse(
                bytes(
                    AugSchemeMPL.aggregate(
                        [
                            AugSchemeMPL.sign(other_sk, test_name, sum_pk),
                            AugSchemeMPL.sign(child_sk, test_name, sum_pk),
                        ]
                    )
                ),
                test_name,
            ),
        ]

    # Test partial signing
    test_name = std_hash(b"path and sum hint partial")
    test_name_2 = std_hash(test_name)
    root_sk_2 = PrivateKey.from_bytes(std_hash(b"a key we do not have"))
    child_sk = _derive_path_unhardened(root_sk, [uint64(1), uint64(2), uint64(3), uint64(4)])
    child_sk_2 = _derive_path_unhardened(root_sk_2, [uint64(1), uint64(2), uint64(3), uint64(4)])
    other_sk = PrivateKey.from_bytes(test_name)
    other_sk_2 = PrivateKey.from_bytes(test_name_2)
    sum_pk = child_sk.get_g1() + other_sk.get_g1()
    sum_pk_2 = child_sk_2.get_g1() + other_sk_2.get_g1()
    signing_responses = await wallet.execute_signing_instructions(
        SigningInstructions(
            KeyHints(
                [
                    SumHint([child_sk.get_g1().get_fingerprint().to_bytes(4, "big")], test_name, bytes(sum_pk)),
                    SumHint([child_sk_2.get_g1().get_fingerprint().to_bytes(4, "big")], test_name_2, bytes(sum_pk_2)),
                ],
                [
                    PathHint(root_fingerprint, [uint64(1), uint64(2), uint64(3), uint64(4)]),
                    PathHint(
                        root_sk_2.get_g1().get_fingerprint().to_bytes(4, "big"),
                        [uint64(1), uint64(2), uint64(3), uint64(4)],
                    ),
                ],
            ),
            [
                SigningTarget(sum_pk.get_fingerprint().to_bytes(4, "big"), test_name, test_name),
                SigningTarget(sum_pk_2.get_fingerprint().to_bytes(4, "big"), test_name_2, test_name_2),
            ],
        ),
        partial_allowed=True,
    )
    assert signing_responses == [
        SigningResponse(
            bytes(
                AugSchemeMPL.sign(child_sk, test_name, sum_pk),
            ),
            test_name,
        ),
        SigningResponse(
            bytes(AugSchemeMPL.sign(other_sk, test_name, sum_pk)),
            test_name,
        ),
        SigningResponse(bytes(AugSchemeMPL.sign(other_sk_2, test_name_2, sum_pk_2)), test_name_2),
    ]

    # Test errors
    unknown_path_hint = SigningInstructions(
        KeyHints(
            [],
            [PathHint(b"unknown fingerprint", [uint64(1), uint64(2), uint64(3), uint64(4)])],
        ),
        [],
    )
    unknown_sum_hint = SigningInstructions(
        KeyHints(
            [SumHint([b"unknown fingerprint"], b"", bytes(G1Element()))],
            [],
        ),
        [],
    )
    unknown_target = SigningInstructions(
        KeyHints(
            [],
            [],
        ),
        [SigningTarget(b"unknown fingerprint", b"", std_hash(b"some hook"))],
    )
    with pytest.raises(ValueError, match="No root pubkey for fingerprint"):
        await wallet.execute_signing_instructions(unknown_path_hint)
    with pytest.raises(ValueError, match="No pubkey found"):
        await wallet.execute_signing_instructions(unknown_sum_hint)
    with pytest.raises(ValueError, match="not found"):
        await wallet.execute_signing_instructions(unknown_target)

    # Test no private key partial sign sum hint
    wallet.wallet_state_manager.private_key = None
    test_name = std_hash(b"sum hint partial no private key")
    other_sk = PrivateKey.from_bytes(test_name)
    sum_pk = other_sk.get_g1() + root_pk
    signing_responses = await wallet.execute_signing_instructions(
        SigningInstructions(
            KeyHints(
                [SumHint([root_fingerprint], test_name, bytes(sum_pk))],
                [],
            ),
            [SigningTarget(sum_pk.get_fingerprint().to_bytes(4, "big"), test_name, test_name)],
        ),
        partial_allowed=True,
    )
    assert signing_responses == [SigningResponse(bytes(AugSchemeMPL.sign(other_sk, test_name, sum_pk)), test_name)]


def test_blind_signer_translation_layer() -> None:
    sum_hints: list[SumHint] = [
        SumHint([b"a", b"b", b"c"], b"offset", b"final"),
        SumHint([b"c", b"b", b"a"], b"offset2", b"final"),
    ]
    path_hints: list[PathHint] = [
        PathHint(b"root1", [uint64(1), uint64(2), uint64(3)]),
        PathHint(b"root2", [uint64(4), uint64(5), uint64(6)]),
    ]
    signing_targets: list[SigningTarget] = [
        SigningTarget(b"pubkey", b"message", bytes32.zeros),
        SigningTarget(b"pubkey2", b"message2", bytes32([1] * 32)),
    ]

    instructions: SigningInstructions = SigningInstructions(
        KeyHints(sum_hints, path_hints),
        signing_targets,
    )
    transaction: UnsignedTransaction = UnsignedTransaction(
        TransactionInfo([]),
        instructions,
    )
    signing_response: SigningResponse = SigningResponse(
        b"signature",
        bytes32([1] * 32),
    )

    bstl_sum_hints: list[BSTLSumHint] = [
        BSTLSumHint([b"a", b"b", b"c"], b"offset", b"final"),
        BSTLSumHint([b"c", b"b", b"a"], b"offset2", b"final"),
    ]
    bstl_path_hints: list[BSTLPathHint] = [
        BSTLPathHint(b"root1", [uint64(1), uint64(2), uint64(3)]),
        BSTLPathHint(b"root2", [uint64(4), uint64(5), uint64(6)]),
    ]
    bstl_signing_targets: list[BSTLSigningTarget] = [
        BSTLSigningTarget(b"pubkey", b"message", bytes32.zeros),
        BSTLSigningTarget(b"pubkey2", b"message2", bytes32([1] * 32)),
    ]

    bstl_instructions: BSTLSigningInstructions = BSTLSigningInstructions(
        bstl_sum_hints,
        bstl_path_hints,
        bstl_signing_targets,
    )
    bstl_transaction: BSTLUnsignedTransaction = BSTLUnsignedTransaction(
        bstl_sum_hints,
        bstl_path_hints,
        bstl_signing_targets,
    )
    bstl_signing_response: BSTLSigningResponse = BSTLSigningResponse(
        b"signature",
        bytes32([1] * 32),
    )
    bstl_instructions_json = json_serialize_with_klvm_streamable(bstl_instructions)
    bstl_transaction_json = json_serialize_with_klvm_streamable(bstl_transaction)
    bstl_signing_response_json = json_serialize_with_klvm_streamable(bstl_signing_response)
    assert bstl_instructions_json == json_serialize_with_klvm_streamable(
        instructions, translation_layer=BLIND_SIGNER_TRANSLATION
    )
    assert bstl_transaction_json == json_serialize_with_klvm_streamable(
        transaction, translation_layer=BLIND_SIGNER_TRANSLATION
    )
    assert bstl_signing_response_json == json_serialize_with_klvm_streamable(
        signing_response, translation_layer=BLIND_SIGNER_TRANSLATION
    )

    assert (
        json_deserialize_with_klvm_streamable(
            bstl_instructions_json, SigningInstructions, translation_layer=BLIND_SIGNER_TRANSLATION
        )
        == instructions
    )
    assert (
        json_deserialize_with_klvm_streamable(
            bstl_transaction_json, UnsignedTransaction, translation_layer=BLIND_SIGNER_TRANSLATION
        )
        == transaction
    )
    assert (
        json_deserialize_with_klvm_streamable(
            bstl_signing_response_json, SigningResponse, translation_layer=BLIND_SIGNER_TRANSLATION
        )
        == signing_response
    )


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [1],
            "trusted": True,
            "reuse_puzhash": True,
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_signer_commands(wallet_environments: WalletTestFramework) -> None:
    wallet: Wallet = wallet_environments.environments[0].xck_wallet
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client
    client_info: WalletClientInfo = WalletClientInfo(
        wallet_rpc,
        wallet_state_manager.root_pubkey.get_fingerprint(),
        wallet_state_manager.config,
    )

    AMOUNT = uint64(1)
    async with wallet_state_manager.new_action_scope(DEFAULT_TX_CONFIG, sign=False, push=False) as action_scope:
        await wallet.generate_signed_transaction(AMOUNT, bytes32.zeros, action_scope)
    [tx] = action_scope.side_effects.transactions

    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("./temp-tb", "wb") as file:
            file.write(bytes(TransactionBundle([tx])))

        await GatherSigningInfoCMD(
            rpc_info=NeedsWalletRPC(client_info=client_info),
            sp_out=SPOut(
                translation="CHIP-0028",
                output_format="file",
                output_file=["./temp-si"],
            ),
            txs_in=TransactionsIn(transaction_file_in="./temp-tb"),
        ).run()

        await ExecuteSigningInstructionsCMD(
            rpc_info=NeedsWalletRPC(client_info=client_info),
            sp_in=SPIn(
                translation="CHIP-0028",
                signer_protocol_input=["./temp-si"],
            ),
            sp_out=SPOut(
                translation="CHIP-0028",
                output_format="file",
                output_file=["./temp-sr"],
            ),
        ).run()

        await ApplySignaturesCMD(
            rpc_info=NeedsWalletRPC(client_info=client_info),
            txs_in=TransactionsIn(transaction_file_in="./temp-tb"),
            sp_in=SPIn(
                translation="CHIP-0028",
                signer_protocol_input=["./temp-sr"],
            ),
            txs_out=TransactionsOut(transaction_file_out="./temp-stb"),
        ).run()

        await PushTransactionsCMD(
            rpc_info=NeedsWalletRPC(client_info=client_info),
            txs_in=TransactionsIn(transaction_file_in="./temp-stb"),
        ).run()

        await wallet_environments.process_pending_states(
            [
                WalletStateTransition(
                    pre_block_balance_updates={
                        1: {
                            "unconfirmed_wallet_balance": -1 * AMOUNT,
                            "<=#spendable_balance": -1 * AMOUNT,
                            "<=#max_send_amount": -1 * AMOUNT,
                            "pending_change": sum(c.amount for c in tx.removals) - AMOUNT,
                            "pending_coin_removal_count": 1,
                        }
                    },
                    post_block_balance_updates={
                        1: {
                            "confirmed_wallet_balance": -1 * AMOUNT,
                            "pending_change": -1 * (sum(c.amount for c in tx.removals) - AMOUNT),
                            "pending_coin_removal_count": -1,
                            "set_remainder": True,
                        },
                    },
                ),
            ]
        )


def test_signer_command_default_parsing() -> None:
    check_click_parsing(
        GatherSigningInfoCMD(
            rpc_info=NeedsWalletRPC(client_info=None, wallet_rpc_port=None, fingerprint=None),
            sp_out=SPOut(
                translation="none",
                output_format="hex",
                output_file=tuple(),
            ),
            txs_in=TransactionsIn(transaction_file_in="in"),
        ),
        "--transaction-file-in",
        "in",
    )

    check_click_parsing(
        ExecuteSigningInstructionsCMD(
            rpc_info=NeedsWalletRPC(client_info=None, wallet_rpc_port=None, fingerprint=None),
            sp_in=SPIn(
                translation="none",
                signer_protocol_input=("sp-in",),
            ),
            sp_out=SPOut(
                translation="none",
                output_format="hex",
                output_file=tuple(),
            ),
        ),
        "-p",
        "sp-in",
    )

    check_click_parsing(
        ApplySignaturesCMD(
            rpc_info=NeedsWalletRPC(client_info=None, wallet_rpc_port=None, fingerprint=None),
            txs_in=TransactionsIn(transaction_file_in="in"),
            sp_in=SPIn(
                translation="none",
                signer_protocol_input=("sp-in",),
            ),
            txs_out=TransactionsOut(transaction_file_out="out"),
        ),
        "--transaction-file-in",
        "in",
        "--transaction-file-out",
        "out",
        "-p",
        "sp-in",
    )

    check_click_parsing(
        PushTransactionsCMD(
            rpc_info=NeedsWalletRPC(client_info=None, wallet_rpc_port=None, fingerprint=None),
            txs_in=TransactionsIn(transaction_file_in="in"),
        ),
        "--transaction-file-in",
        "in",
    )


def test_transactions_in() -> None:
    @click.group()
    def cmd() -> None:
        pass

    @chik_command(group=cmd, name="temp_cmd", short_help="blah", help="n/a")
    class TempCMD(TransactionsIn):
        def run(self) -> None:
            assert self.transaction_bundle == TransactionBundle([STD_TX])

    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("some file", "wb") as file:
            file.write(bytes(TransactionBundle([STD_TX])))

        result = runner.invoke(cmd, ["temp_cmd", "--transaction-file-in", "some file"], catch_exceptions=False)
        assert result.output == ""


def test_transactions_out() -> None:
    @click.group()
    def cmd() -> None:
        pass

    @chik_command(group=cmd, name="temp_cmd", short_help="blah", help="n/a")
    class TempCMD(TransactionsOut):
        def run(self) -> None:
            self.handle_transaction_output([STD_TX])

    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(cmd, ["temp_cmd", "--transaction-file-out", "some file"], catch_exceptions=False)
        assert result.output == ""

        with open("some file", "rb") as file:
            file.read() == bytes(TransactionBundle([STD_TX]))


@klvm_streamable
@dataclasses.dataclass(frozen=True)
class FooCoin(Streamable):
    amount: uint64

    @staticmethod
    def from_wallet_api(_from: Coin) -> FooCoin:
        return FooCoin(_from.amount)

    @staticmethod
    def to_wallet_api(_from: FooCoin) -> Coin:
        return Coin(
            bytes32.zeros,
            bytes32.zeros,
            _from.amount,
        )


FOO_COIN_TRANSLATION = TranslationLayer(
    [
        TranslationLayerMapping(
            Coin,
            FooCoin,
            FooCoin.from_wallet_api,
            FooCoin.to_wallet_api,
        )
    ]
)


def test_signer_protocol_in(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setitem(ALL_TRANSLATION_LAYERS, "CHIP-0028", FOO_COIN_TRANSLATION)

    @click.group()
    def cmd() -> None:
        pass

    coin = Coin(bytes32.zeros, bytes32.zeros, uint64(13))

    @chik_command(group=cmd, name="temp_cmd", short_help="blah", help="n/a")
    class TempCMD(SPIn):
        def run(self) -> None:
            assert self.read_sp_input(Coin) == [coin, coin]

    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("some file", "wb") as file:
            file.write(byte_serialize_klvm_streamable(coin))

        with open("some file2", "wb") as file:
            file.write(byte_serialize_klvm_streamable(coin))

        result = runner.invoke(
            cmd,
            ["temp_cmd", "--signer-protocol-input", "some file", "--signer-protocol-input", "some file2"],
            catch_exceptions=False,
        )
        assert result.output == ""

    with runner.isolated_filesystem():
        with open("some file", "wb") as file:
            file.write(byte_serialize_klvm_streamable(coin, translation_layer=FOO_COIN_TRANSLATION))

            with open("some file2", "wb") as file:
                file.write(byte_serialize_klvm_streamable(coin, translation_layer=FOO_COIN_TRANSLATION))

        result = runner.invoke(
            cmd, ["temp_cmd", "--signer-protocol-input", "some file", "--signer-protocol-input", "some file2"]
        )
        assert result.exception is not None
        result = runner.invoke(
            cmd,
            [
                "temp_cmd",
                "--signer-protocol-input",
                "some file",
                "--signer-protocol-input",
                "some file2",
                "--translation",
                "CHIP-0028",
            ],
            catch_exceptions=False,
        )
        assert result.output == ""


def test_signer_protocol_out(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setitem(ALL_TRANSLATION_LAYERS, "CHIP-0028", FOO_COIN_TRANSLATION)

    @click.group()
    def cmd() -> None:
        pass

    coin = Coin(bytes32.zeros, bytes32.zeros, uint64(0))
    coin_bytes = byte_serialize_klvm_streamable(coin)

    @chik_command(group=cmd, name="temp_cmd", short_help="blah", help="n/a")
    class TempCMD(SPOut):
        def run(self) -> None:
            self.handle_klvm_output([coin, coin])

    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(cmd, ["temp_cmd", "--output-format", "hex"], catch_exceptions=False)
        assert result.output.strip() == coin_bytes.hex() + "\n" + coin_bytes.hex()

        result = runner.invoke(cmd, ["temp_cmd", "--output-format", "file"], catch_exceptions=False)
        assert result.output == "--output-format=file specified without any --output-file\n"

        result = runner.invoke(
            cmd, ["temp_cmd", "--output-format", "file", "--output-file", "some file"], catch_exceptions=False
        )
        assert "Incorrect number of file outputs specified" in result.output

        result = runner.invoke(
            cmd,
            ["temp_cmd", "--output-format", "file", "--output-file", "some file", "--output-file", "some file2"],
            catch_exceptions=False,
        )
        assert result.output == ""

        with open("some file", "rb") as file:
            file.read() == coin_bytes

        with open("some file2", "rb") as file:
            file.read() == coin_bytes

        result = runner.invoke(cmd, ["temp_cmd", "--output-format", "qr"], catch_exceptions=False)
        assert result.output != ""  # separate test for QrCodeDisplay

        result = runner.invoke(
            cmd, ["temp_cmd", "--output-format", "hex", "--translation", "CHIP-0028"], catch_exceptions=False
        )
        assert result.output.strip() != coin_bytes.hex()
        coin_hex = byte_serialize_klvm_streamable(coin, translation_layer=ALL_TRANSLATION_LAYERS["CHIP-0028"]).hex()
        assert result.output.strip() == coin_hex + "\n" + coin_hex


def test_qr_code_display() -> None:
    @click.group()
    def cmd() -> None:
        pass

    bytes_to_encode = b"foo bar qat qux bam bat"

    @chik_command(group=cmd, name="temp_cmd", short_help="blah", help="n/a")
    class TempCMD(QrCodeDisplay):
        def run(self) -> None:
            self.display_qr_codes([bytes_to_encode, bytes_to_encode])

    runner = CliRunner()
    result = runner.invoke(
        cmd,
        ["temp_cmd"],
        input="\n",
        catch_exceptions=False,
    )

    # Would be good to check eventually that the QR codes are valid but segno doesn't seem to provide that ATM
    assert result.output.count("Displaying QR Codes (1/2)") == 1
    assert result.output.count("Displaying QR Codes (2/2)") == 1



from __future__ import annotations

from collections.abc import Generator, Iterable, KeysView

SERVICES_FOR_GROUP: dict[str, list[str]] = {
    "all": [
        "chik_harvester",
        "chik_timelord_launcher",
        "chik_timelord",
        "chik_farmer",
        "chik_full_node",
        "chik_wallet",
        "chik_data_layer",
        "chik_data_layer_http",
    ],
    "daemon": [],
    # TODO: should this be `data_layer`?
    "data": ["chik_wallet", "chik_data_layer"],
    "data_layer_http": ["chik_data_layer_http"],
    "node": ["chik_full_node"],
    "harvester": ["chik_harvester"],
    "farmer": ["chik_harvester", "chik_farmer", "chik_full_node", "chik_wallet"],
    "farmer-no-wallet": ["chik_harvester", "chik_farmer", "chik_full_node"],
    "farmer-only": ["chik_farmer"],
    "timelord": ["chik_timelord_launcher", "chik_timelord", "chik_full_node"],
    "timelord-only": ["chik_timelord"],
    "timelord-launcher-only": ["chik_timelord_launcher"],
    "wallet": ["chik_wallet"],
    "introducer": ["chik_introducer"],
    "simulator": ["chik_full_node_simulator"],
    "crawler": ["chik_crawler"],
    "seeder": ["chik_crawler", "chik_seeder"],
    "seeder-only": ["chik_seeder"],
}


def all_groups() -> KeysView[str]:
    return SERVICES_FOR_GROUP.keys()


def services_for_groups(groups: Iterable[str]) -> Generator[str, None, None]:
    for group in groups:
        yield from SERVICES_FOR_GROUP[group]


def validate_service(service: str) -> bool:
    return any(service in _ for _ in SERVICES_FOR_GROUP.values())



from __future__ import annotations

import textwrap
from pathlib import Path
from typing import Any, Callable

import click
import pytest
import yaml
from click.testing import CliRunner

from chik.util.virtual_project_analysis import (
    Annotation,
    ChikFile,
    Config,
    DirectoryParameters,
    File,
    Package,
    build_dependency_graph,
    build_virtual_dependency_graph,
    cli,
    config,
    find_cycles,
    parse_file_or_package,
)


@pytest.mark.parametrize(
    "file_string, expected, annotated",
    [
        ("# Package: example\n# Some other comment", "example", True),
        ("# Some comment\n# Some other comment", "chik-blockchain", False),
    ],
)
def test_parse_annotation(file_string: str, expected: str, annotated: bool) -> None:
    """
    Test that parse returns an Annotation instance for a valid annotation or
    raises ValueError for an invalid one.
    """
    annotation = Annotation.parse(file_string)
    assert isinstance(annotation, Annotation)
    assert annotation.package == expected
    assert annotation.is_annotated == annotated


# Temporary directory fixture to create test files
@pytest.fixture
def create_test_file(tmp_path: Path) -> Callable[[str, str], Path]:
    def _create_test_file(name: str, content: str) -> Path:
        file_path = tmp_path / name
        file_path.write_text(content, encoding="utf-8")
        return file_path

    return _create_test_file


def test_parse_with_annotation(create_test_file: Callable[[str, str], Path]) -> None:
    """Test parsing a file that contains a valid annotation."""
    file_content = "# Package: test_package\n# Some other comment"
    test_file = create_test_file("annotated_file.txt", file_content)

    parsed_file = ChikFile.parse(test_file)

    assert parsed_file.path == test_file
    assert isinstance(parsed_file.annotations, Annotation)
    assert parsed_file.annotations.package == "test_package"


def test_parse_without_annotation(create_test_file: Callable[[str, str], Path]) -> None:
    """Test parsing a file that does not contain any annotations."""
    file_content = "# Some comment\n# Some other comment"
    test_file = create_test_file("non_annotated_file.txt", file_content)

    parsed_file = ChikFile.parse(test_file)

    assert parsed_file.path == test_file
    assert not parsed_file.annotations.is_annotated


# This test is optional and can be adapted based on expected behavior for non-existent files
def test_parse_nonexistent_file() -> None:
    """Test attempting to parse a non-existent file."""
    with pytest.raises(FileNotFoundError):
        ChikFile.parse(Path("/path/to/nonexistent/file.txt"))


# Helper function to create a non-empty Python file
def create_python_file(dir_path: Path, name: str, content: str) -> Path:
    file_path = dir_path / name
    file_path.write_text(content, encoding="utf-8")
    return file_path


# Helper function to create an empty Python file
def create_empty_python_file(dir_path: Path, name: str) -> Path:
    file_path = dir_path / name
    file_path.touch()
    return file_path


def test_gather_non_empty_python_files(tmp_path: Path) -> None:
    # Set up directory structure
    dir_path = tmp_path / "test_dir"
    dir_path.mkdir()
    excluded_dir = tmp_path / "excluded_dir"
    excluded_dir.mkdir()

    # Create test files
    non_empty_file = create_python_file(dir_path, "non_empty.py", "print('Hello World')")
    create_empty_python_file(dir_path, "empty.py")
    create_python_file(excluded_dir, "excluded.py", "print('Hello World')")

    # Initialize DirectoryParameters with excluded paths
    dir_params = DirectoryParameters(dir_path=dir_path, excluded_paths=[excluded_dir])

    # Perform the test
    python_files = dir_params.gather_non_empty_python_files()

    # Assertions
    assert len(python_files) == 1  # Only one non-empty Python file should be found
    assert python_files[0].path == non_empty_file  # The path of the gathered file should match the non-empty file


def test_gather_with_nested_directories_and_exclusions(tmp_path: Path) -> None:
    # Set up directory structure
    base_dir = tmp_path / "base_dir"
    base_dir.mkdir()
    nested_dir = base_dir / "nested_dir"
    nested_dir.mkdir()
    excluded_dir = base_dir / "excluded_dir"
    excluded_dir.mkdir()

    # Create test files
    nested_file = create_python_file(nested_dir, "nested.py", "print('Hello World')")
    create_empty_python_file(nested_dir, "nested_empty.py")
    create_python_file(excluded_dir, "excluded.py", "print('Hello World')")

    # Initialize DirectoryParameters without excluded paths
    dir_params = DirectoryParameters(dir_path=base_dir, excluded_paths=[excluded_dir])

    # Perform the test
    python_files = dir_params.gather_non_empty_python_files()

    # Assertions
    assert len(python_files) == 1  # Only the non-empty Python file in the nested directory should be found
    assert python_files[0].path == nested_file  # The path of the gathered file should match the nested non-empty file


def test_find_missing_annotations(tmp_path: Path) -> None:
    # Set up directory structure
    dir_path = tmp_path / "test_dir"
    dir_path.mkdir()

    # Create test files
    create_python_file(dir_path, "non_empty.py", "print('Hello World')")

    # Run the command
    runner = CliRunner()
    result = runner.invoke(cli, ["find_missing_annotations", "--directory", str(dir_path)])
    assert result.output == f"{dir_path / 'non_empty.py'}\n"

    # Rewrite file to have annotation
    create_python_file(dir_path, "non_empty.py", "# Package: misc\n")

    # Run the command again with no results
    result = runner.invoke(cli, ["find_missing_annotations", "--directory", str(dir_path)])
    assert result.output == ""


def test_parse_file_or_package() -> None:
    assert parse_file_or_package("example.py") == File(Path("example.py"))
    assert parse_file_or_package("example.py (extra info)") == File(Path("example.py"))
    assert parse_file_or_package("(package_name)") == Package("package_name")
    assert parse_file_or_package("package_name") == Package("package_name")
    assert parse_file_or_package("package_name(") == Package("package_name(")
    assert parse_file_or_package("(package_name") == Package("(package_name")
    assert parse_file_or_package("package_name)") == Package("package_name)")


@pytest.fixture
def chik_package_structure(tmp_path: Path) -> Path:
    base_dir = tmp_path / "chik_project"
    base_dir.mkdir()
    chik_dir = base_dir / "chik"
    chik_dir.mkdir()

    # Create some files within the chik package
    create_python_file(chik_dir, "module1.py", "def func1(): pass")
    create_python_file(chik_dir, "module2.py", "def func2(): pass\nfrom chik.module1 import func1")
    create_python_file(chik_dir, "module3.py", "def func3(): pass\nimport chik.module2")

    return chik_dir


def test_build_dependency_graph(chik_package_structure: Path) -> None:
    chik_dir = chik_package_structure
    dir_params = DirectoryParameters(dir_path=chik_dir)
    graph = build_dependency_graph(dir_params)
    assert chik_dir / "module1.py" in graph
    assert chik_dir / "module2.py" in graph
    assert chik_dir / "module3.py" in graph
    assert chik_dir / "module1.py" in graph[chik_dir / "module2.py"]
    assert chik_dir / "module2.py" in graph[chik_dir / "module3.py"]


def test_print_dependency_graph(chik_package_structure: Path) -> None:
    # Run the command
    runner = CliRunner()
    result = runner.invoke(cli, ["print_dependency_graph", "--directory", str(chik_package_structure)])
    assert "module1.py" in result.output
    assert "module2.py" in result.output
    assert "module3.py" in result.output


# Mock the build_dependency_graph function to control its output
def mock_build_dependency_graph(dir_params: DirectoryParameters) -> dict[Path, list[Path]]:
    return {
        Path("/path/to/package1/module1.py"): [
            Path("/path/to/package2/module2.py"),
            Path("/path/to/package3/module3.py"),
        ],
        Path("/path/to/package2/module2.py"): [],
        Path("/path/to/package3/module3.py"): [Path("/path/to/package2/module2.py")],
    }


# Helper function to simulate ChikFile.parse for testing
def mock_chik_file_parse(path: Path) -> ChikFile:
    annotations_map = {
        Path("/path/to/package1/module1.py"): Annotation("Package1", True),
        Path("/path/to/package2/module2.py"): Annotation("Package2", True),
        Path("/path/to/package3/module3.py"): Annotation("Package3", True),
    }
    return ChikFile(path=Path(path), annotations=annotations_map.get(path, Annotation("chik-blockchain", False)))


@pytest.fixture
def prepare_mocks(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("chik.util.virtual_project_analysis.build_dependency_graph", mock_build_dependency_graph)
    monkeypatch.setattr("chik.util.virtual_project_analysis.ChikFile.parse", mock_chik_file_parse)


def test_build_virtual_dependency_graph(prepare_mocks: None) -> None:
    dir_params = DirectoryParameters(dir_path=Path("/path/to/package1"))
    virtual_graph = build_virtual_dependency_graph(dir_params)

    assert "Package2" in virtual_graph["Package1"]
    assert "Package3" in virtual_graph["Package1"]
    assert virtual_graph["Package2"] == []
    assert "Package2" in virtual_graph["Package3"]


def test_print_virtual_dependency_graph(tmp_path: Path) -> None:
    chik_dir = tmp_path / "chik"
    chik_dir.mkdir()

    # Create some files within the chik package
    create_python_file(chik_dir, "module1.py", "# Package: one\ndef func1(): pass")
    create_python_file(chik_dir, "module2.py", "# Package: two\ndef func2(): pass\nfrom chik.module1 import func1")
    create_python_file(chik_dir, "module3.py", "# Package: three\ndef func3(): pass\nimport chik.module2")

    # Run the command
    runner = CliRunner()
    result = runner.invoke(cli, ["print_virtual_dependency_graph", "--directory", str(chik_dir)])
    assert "one" in result.output
    assert "two" in result.output
    assert "three" in result.output


# Helper function to simulate ChikFile.parse for testing
def mock_chik_file_parse2(path: Path) -> ChikFile:
    annotations_map = {
        Path("/path/to/package1/module1.py"): Annotation("Package1", True),
        Path("/path/to/package2/module2.py"): Annotation("Package2", True),
        Path("/path/to/package3/module3.py"): Annotation("Package1", True),
    }
    return ChikFile(path=Path(path), annotations=annotations_map.get(path, Annotation("chik-blockchain", False)))


@pytest.fixture
def prepare_mocks2(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("chik.util.virtual_project_analysis.ChikFile.parse", mock_chik_file_parse2)


def test_cycle_detection(prepare_mocks2: None) -> None:
    # Example graph with a simple cycle
    graph: dict[Path, list[Path]] = {
        Path("/path/to/package1/module1.py"): [Path("/path/to/package2/module2.py")],
        Path("/path/to/package2/module2.py"): [Path("/path/to/package3/module3.py")],  # Cycle here
        Path("/path/to/package3/module3.py"): [],
    }
    cycles = find_cycles(
        graph,
        build_virtual_dependency_graph(DirectoryParameters(dir_path=Path("path")), existing_graph=graph),
        excluded_paths=[],
        ignore_cycles_in=[],
        ignore_specific_files=[],
        ignore_specific_edges=[],
    )
    # \path\to\package1\module1.py (Package1) -> \path\to\package2\module2.py (Package2) -> (Package1)
    # \path\to\package2\module2.py (Package2) -> \path\to\package3\module3.py (Package1) -> (Package2)
    assert len(cycles) == 2


def test_print_cycles(tmp_path: Path) -> None:
    chik_dir = tmp_path / "chik"
    chik_dir.mkdir()

    # Create some files within the chik package
    create_python_file(chik_dir, "module1.py", "# Package: one\ndef func1(): pass\nfrom chik.module2 import func2")
    create_python_file(chik_dir, "module2.py", "# Package: two\ndef func2(): pass\nfrom chik.module3 import func3")
    create_python_file(chik_dir, "module3.py", "# Package: one\ndef func3(): pass\n")

    # Run the command
    runner = CliRunner()
    result = runner.invoke(cli, ["print_cycles", "--directory", str(chik_dir)])
    assert "module1.py (one) -> " in result.output


def test_check_config(tmp_path: Path) -> None:
    chik_dir = tmp_path / "chik"
    chik_dir.mkdir()

    # Create some files within the chik package
    create_python_file(
        chik_dir,
        "module1.py",
        textwrap.dedent(
            """
            # Package: one
            def func1(): pass
            from chik.module2 import func2
            """
        ),
    )
    create_python_file(
        chik_dir,
        "module1b.py",
        textwrap.dedent(
            """
            # Package: one
            def func1b(): pass
            """
        ),
    )
    create_python_file(
        chik_dir,
        "module2.py",
        textwrap.dedent(
            """
            # Package: two
            def func2(): pass
            from chik.module3 import func3
            from chik.module1b import func1b
            """
        ),
    )
    create_python_file(
        chik_dir,
        "module3.py",
        textwrap.dedent(
            """
            # Package: three
            def func3(): pass
            """
        ),
    )

    # Run the command
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "check_config",
            "--directory",
            str(chik_dir),
            "--ignore-cycles-in",
            "three",
            "--ignore-specific-file",
            str(chik_dir / "module3.py"),
            "--ignore-specific-edge",
            str(chik_dir / "module2.py") + " -> " + str(chik_dir / "module3.py"),
            "--ignore-specific-edge",
            str(chik_dir / "module2.py") + " -> " + str(chik_dir / "module1b.py"),
        ],
    )
    assert "    module three ignored but no cycles were found" in result.output
    assert f"    file {chik_dir / 'module3.py'!s} ignored but no cycles were found" in result.output
    assert (
        f"edge {str(chik_dir / 'module2.py') + ' -> ' + str(chik_dir / 'module3.py')} ignored but no cycles were found"
        in result.output
    )
    assert (
        f"edge {str(chik_dir / 'module2.py') + ' -> ' + str(chik_dir / 'module1b.py')} ignored but no cycles were found"
        not in result.output
    )


def test_excluded_paths_handling(prepare_mocks2: None) -> None:
    # Graph where module2.py is excluded
    graph = {
        Path("/path/to/package1/module1.py"): [Path("/path/to/package2/module2.py")],
        Path("/path/to/package2/module2.py"): [Path("/path/to/package1/module1.py")],
    }
    cycles = find_cycles(
        graph,
        build_virtual_dependency_graph(DirectoryParameters(dir_path=Path("path")), existing_graph=graph),
        excluded_paths=[Path("/path/to/package2/module2.py")],
        ignore_cycles_in=[],
        ignore_specific_files=[],
        ignore_specific_edges=[],
    )
    assert len(cycles) == 0  # No cycles due to exclusion


def test_ignore_cycles_in_specific_packages(prepare_mocks2: None) -> None:
    graph: dict[Path, list[Path]] = {
        Path("/path/to/package1/module1.py"): [Path("/path/to/package2/module2.py")],
        Path("/path/to/package2/module2.py"): [Path("/path/to/package3/module3.py")],
        Path("/path/to/package3/module3.py"): [],
    }
    # Assuming module1.py and module3.py belong to Package1, which is ignored
    cycles = find_cycles(
        graph,
        build_virtual_dependency_graph(DirectoryParameters(dir_path=Path("path")), existing_graph=graph),
        excluded_paths=[],
        ignore_cycles_in=["Package1"],
        ignore_specific_files=[],
        ignore_specific_edges=[],
    )
    assert len(cycles) == 1  # Cycles in Package1 are ignored


def test_ignore_cycles_with_specific_edges(monkeypatch: pytest.MonkeyPatch) -> None:
    def _mock_chik_file_parse(path: Path) -> ChikFile:
        annotations_map = {
            Path("/path/to/package1/module1a.py"): Annotation("Package1", True),
            Path("/path/to/package2/module2.py"): Annotation("Package2", True),
            Path("/path/to/package3/module3.py"): Annotation("Package3", True),
            Path("/path/to/package1/module1b.py"): Annotation("Package1", True),
        }
        return ChikFile(path=Path(path), annotations=annotations_map.get(path, Annotation("chik-blockchain", False)))

    monkeypatch.setattr("chik.util.virtual_project_analysis.ChikFile.parse", _mock_chik_file_parse)

    graph = {
        Path("/path/to/package1/module1a.py"): [Path("/path/to/package2/module2.py")],
        Path("/path/to/package2/module2.py"): [Path("/path/to/package3/module3.py")],
        Path("/path/to/package3/module3.py"): [Path("/path/to/package1/module1b.py")],
        Path("/path/to/package1/module1b.py"): [],
    }
    virtual_graph = build_virtual_dependency_graph(DirectoryParameters(dir_path=Path("path")), existing_graph=graph)
    cycles = find_cycles(
        graph,
        virtual_graph,
        excluded_paths=[],
        ignore_cycles_in=[],
        ignore_specific_files=[],
        ignore_specific_edges=[
            (File(Path("/path/to/package3/module3.py")), File(Path("/path/to/package2/module2.py"))),
            (Package("Package3"), Package("Package2")),
        ],
    )
    assert len(cycles) == 0


def test_ignore_cycles_with_specific_files(prepare_mocks2: None) -> None:
    graph = {
        Path("/path/to/package1/module1.py"): [Path("/path/to/package2/module2.py")],
        Path("/path/to/package2/module2.py"): [Path("/path/to/package3/module3.py")],  # Cycle here
    }
    cycles = find_cycles(
        graph,
        build_virtual_dependency_graph(DirectoryParameters(dir_path=Path("path")), existing_graph=graph),
        excluded_paths=[],
        ignore_cycles_in=[],
        ignore_specific_files=[Path("/path/to/package1/module1.py"), Path("/path/to/package2/module2.py")],
        ignore_specific_edges=[],
    )
    assert len(cycles) == 0


# Sample function to use with the decorator for testing
@click.command("blah")
@config
def sample_function(config: Config) -> None:
    print(config)


# Helper function to create a temporary YAML configuration file
@pytest.fixture
def create_yaml_config(tmp_path: Path) -> Callable[[dict[str, Any]], Path]:
    def _create_yaml_config(content: dict[str, Any]) -> Path:
        path = tmp_path / "config.yaml"
        with open(path, "w") as f:
            yaml.dump(content, f)
        return path

    return _create_yaml_config


def test_config_with_yaml(create_yaml_config: Callable[[dict[str, Any]], Path]) -> None:
    # Create a temporary YAML configuration file
    yaml_config = {
        "exclude_paths": ["path/to/exclude"],
        "ignore": {
            "packages": ["ignored.package"],
            "files": ["ignored_file.py"],
            "edges": ["ignored_parent -> ignored_child"],
        },
    }
    config_path = create_yaml_config(yaml_config)

    runner = CliRunner()

    # Invoke the CLI with the --config option
    result = runner.invoke(sample_function, ["--directory", ".", "--config", str(config_path)])

    # Check if the command ran successfully
    assert result.exit_code == 0

    # Verify the config object created by the decorator
    config = result.output
    path_type = config_path.__class__.__name__
    assert config == (
        "Config("
        "directory_parameters=DirectoryParameters("
        f"dir_path={path_type}('.'), "
        f"excluded_paths=[{path_type}('path/to/exclude')]"
        "), "
        "ignore_cycles_in=['ignored.package'], "
        f"ignore_specific_files=[{path_type}('ignored_file.py')], "
        f"ignore_specific_edges=[(Package(name='ignored_child', is_file=False), "
        f"Package(name='ignored_parent', is_file=False))]"
        ")\n"
    )


def test_parse_edges(tmp_path: Path) -> None:
    chik_dir = tmp_path / "chik"
    chik_dir.mkdir()

    # Create some files within the chik package
    create_python_file(
        chik_dir,
        "module1.py",
        textwrap.dedent(
            """
            # Package: one
            def func1(): pass
            from chik.module2 import func2
            from chik.module3 import func3
            """
        ),
    )
    create_python_file(
        chik_dir,
        "module2.py",
        textwrap.dedent(
            """
            # Package: two
            def func2(): pass
            """
        ),
    )
    create_python_file(
        chik_dir,
        "module3.py",
        textwrap.dedent(
            """
            # Package: three
            def func3(): pass
            """
        ),
    )

    # Run the command
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "print_edges",
            "--directory",
            str(chik_dir),
            "--dependent-package",
            "one",
            "--provider-package",
            "two",
        ],
    )
    assert result.output.strip() == f"{chik_dir / 'module1.py'!s} (one) -> {chik_dir / 'module2.py'!s} (two)"



from __future__ import annotations

import logging
from typing import Optional

from chik_rs import (
    get_flags_for_height_and_constants,
    run_chik_program,
)
from chik_rs import get_puzzle_and_solution_for_coin2 as get_puzzle_and_solution_for_coin_rust

from chik.consensus.constants import ConsensusConstants
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_record import CoinRecord
from chik.types.coin_spend import CoinSpend, CoinSpendWithConditions, SpendInfo, make_spend
from chik.types.generator_types import BlockGenerator
from chik.types.spend_bundle_conditions import SpendBundleConditions
from chik.util.condition_tools import conditions_for_solution
from chik.util.errors import Err
from chik.util.ints import uint32, uint64
from chik.wallet.puzzles.load_klvm import load_serialized_klvm_maybe_recompile

DESERIALIZE_MOD = load_serialized_klvm_maybe_recompile(
    "chiklisp_deserialisation.clsp", package_or_requirement="chik.consensus.puzzles"
)

log = logging.getLogger(__name__)


def get_puzzle_and_solution_for_coin(
    generator: BlockGenerator, coin: Coin, height: int, constants: ConsensusConstants
) -> SpendInfo:
    try:
        puzzle, solution = get_puzzle_and_solution_for_coin_rust(
            generator.program,
            generator.generator_refs,
            constants.MAX_BLOCK_COST_KLVM,
            coin,
            get_flags_for_height_and_constants(height, constants),
        )
        return SpendInfo(puzzle, solution)
    except Exception as e:
        raise ValueError(f"Failed to get puzzle and solution for coin {coin}, error: {e}") from e


def get_spends_for_block(generator: BlockGenerator, height: int, constants: ConsensusConstants) -> list[CoinSpend]:
    args = bytearray(b"\xff")
    args += bytes(DESERIALIZE_MOD)
    args += b"\xff"
    args += bytes(Program.to(generator.generator_refs))
    args += b"\x80\x80"

    _, ret = run_chik_program(
        bytes(generator.program),
        bytes(args),
        constants.MAX_BLOCK_COST_KLVM,
        get_flags_for_height_and_constants(height, constants),
    )

    spends: list[CoinSpend] = []

    for spend in Program.to(ret).first().as_iter():
        parent, puzzle, amount, solution = spend.as_iter()
        puzzle_hash = puzzle.get_tree_hash()
        coin = Coin(parent.as_atom(), puzzle_hash, uint64(amount.as_int()))
        spends.append(make_spend(coin, puzzle, solution))

    return spends


def get_spends_for_block_with_conditions(
    generator: BlockGenerator, height: int, constants: ConsensusConstants
) -> list[CoinSpendWithConditions]:
    args = bytearray(b"\xff")
    args += bytes(DESERIALIZE_MOD)
    args += b"\xff"
    args += bytes(Program.to(generator.generator_refs))
    args += b"\x80\x80"

    flags = get_flags_for_height_and_constants(height, constants)

    _, ret = run_chik_program(
        bytes(generator.program),
        bytes(args),
        constants.MAX_BLOCK_COST_KLVM,
        flags,
    )

    spends: list[CoinSpendWithConditions] = []

    for spend in Program.to(ret).first().as_iter():
        parent, puzzle, amount, solution = spend.as_iter()
        puzzle_hash = puzzle.get_tree_hash()
        coin = Coin(parent.as_atom(), puzzle_hash, uint64(amount.as_int()))
        coin_spend = make_spend(coin, puzzle, solution)
        conditions = conditions_for_solution(puzzle, solution, constants.MAX_BLOCK_COST_KLVM)
        spends.append(CoinSpendWithConditions(coin_spend, conditions))

    return spends


def mempool_check_time_locks(
    removal_coin_records: dict[bytes32, CoinRecord],
    bundle_conds: SpendBundleConditions,
    prev_transaction_block_height: uint32,
    timestamp: uint64,
) -> Optional[Err]:
    """
    Check all time and height conditions against current state.
    """

    if prev_transaction_block_height < bundle_conds.height_absolute:
        return Err.ASSERT_HEIGHT_ABSOLUTE_FAILED
    if timestamp < bundle_conds.seconds_absolute:
        return Err.ASSERT_SECONDS_ABSOLUTE_FAILED
    if bundle_conds.before_height_absolute is not None:
        if prev_transaction_block_height >= bundle_conds.before_height_absolute:
            return Err.ASSERT_BEFORE_HEIGHT_ABSOLUTE_FAILED
    if bundle_conds.before_seconds_absolute is not None:
        if timestamp >= bundle_conds.before_seconds_absolute:
            return Err.ASSERT_BEFORE_SECONDS_ABSOLUTE_FAILED

    for spend in bundle_conds.spends:
        unspent = removal_coin_records[bytes32(spend.coin_id)]
        if spend.birth_height is not None:
            if spend.birth_height != unspent.confirmed_block_index:
                return Err.ASSERT_MY_BIRTH_HEIGHT_FAILED
        if spend.birth_seconds is not None:
            if spend.birth_seconds != unspent.timestamp:
                return Err.ASSERT_MY_BIRTH_SECONDS_FAILED
        if spend.height_relative is not None:
            if prev_transaction_block_height < unspent.confirmed_block_index + spend.height_relative:
                return Err.ASSERT_HEIGHT_RELATIVE_FAILED
        if spend.seconds_relative is not None:
            if timestamp < unspent.timestamp + spend.seconds_relative:
                return Err.ASSERT_SECONDS_RELATIVE_FAILED
        if spend.before_height_relative is not None:
            if prev_transaction_block_height >= unspent.confirmed_block_index + spend.before_height_relative:
                return Err.ASSERT_BEFORE_HEIGHT_RELATIVE_FAILED
        if spend.before_seconds_relative is not None:
            if timestamp >= unspent.timestamp + spend.before_seconds_relative:
                return Err.ASSERT_BEFORE_SECONDS_RELATIVE_FAILED

    return None



# Package: utils

from __future__ import annotations


def clamp(n: int, smallest: int, largest: int) -> int:
    return max(smallest, min(n, largest))


def make_monotonically_decreasing(seq: list[float]) -> list[float]:
    out: list[float] = []
    if len(seq) > 0:
        min = seq[0]
        for n in seq:
            if n <= min:
                out.append(n)
                min = n
            else:
                out.append(min)
    return out



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint32
from chik.util.streamable import Streamable, streamable


@streamable
@dataclass(frozen=True)
class FarmNewBlockProtocol(Streamable):
    puzzle_hash: bytes32


@streamable
@dataclass(frozen=True)
class ReorgProtocol(Streamable):
    old_index: uint32
    new_index: uint32
    puzzle_hash: bytes32
    seed: Optional[bytes32]


@streamable
@dataclass(frozen=True)
class GetAllCoinsProtocol(Streamable):
    include_spent_coins: bool



from __future__ import annotations

SPACEBUCKS = {
    "asset_id": "a628c1c2c6fcb74d53746157e438e108eab5c0bb3e5c80ff9b1910b3e4832913",
    "name": "Spacebucks",
    "symbol": "SBX",
}

MARMOT = {
    "asset_id": "8ebf855de6eb146db5602f0456d2f0cbe750d57f821b6f91a8592ee9f1d4cf31",
    "name": "Marmot",
    "symbol": "MRMT",
}

STABLY_USDS = {
    "asset_id": "6d95dae356e32a71db5ddcb42224754a02524c615c5fc35f568c2af04774e589",
    "name": "Stably USD Classic",
    "symbol": "USDSC",
}

CHIK_HOLIDAY_TOKEN = {
    "asset_id": "509deafe3cd8bbfbb9ccce1d930e3d7b57b40c964fa33379b18d628175eb7a8f",
    "name": "Chik Holiday 2023 Token",
    "symbol": "CK23",
}

BASE_WARPED_USDC = {
    "asset_id": "fa4a180ac326e67ea289b869e3448256f6af05721f7cf934cb9901baa6b7a99d",
    "name": "Base Warped USDC ",
    "symbol": "wUSDC.b",
}

BASE_WARPED_MILLIETH = {
    "asset_id": "f322a205c034fe28681829fa5a2e483ac421f0952eb1292945c8db06e0a471a6",
    "name": "Base Warped milliETH",
    "symbol": "wmilliETH.b",
}

BASE_WARPED_EURC = {
    "asset_id": "31ffd54c5b38bb33352dc0be0ebf9cf4f29a6329156dd4811d6f4254f85c8200",
    "name": "Base Warped EURC",
    "symbol": "wEURC.b",
}


ETH_WARPED_USDC = {
    "asset_id": "bbb51b246fbec1da1305be31dcf17151ccd0b8231a1ec306d7ce9f5b8c742b9e",
    "name": "Ethereum Warped USDC",
    "symbol": "wUSDC",
}

ETH_WARPED_MILLIETH = {
    "asset_id": "4cb15a8ecc85068fb1f98c09a5e489d1ad61b2af79690ce00f9fc4803c8b597f",
    "name": "Ethereum Warped milliETH",
    "symbol": "wmilliETH",
}

ETH_WARPED_USDT = {
    "asset_id": "634f9f0de1a6c39a2189948b8e61b6852fbf774f73b0e36e143e841c49a0798c",
    "name": "Ethereum Warped USDT",
    "symbol": "wUSDT",
}

DEFAULT_CATS = {
    SPACEBUCKS["asset_id"]: SPACEBUCKS,
    MARMOT["asset_id"]: MARMOT,
    STABLY_USDS["asset_id"]: STABLY_USDS,
    CHIK_HOLIDAY_TOKEN["asset_id"]: CHIK_HOLIDAY_TOKEN,
    BASE_WARPED_USDC["asset_id"]: BASE_WARPED_USDC,
    BASE_WARPED_MILLIETH["asset_id"]: BASE_WARPED_MILLIETH,
    BASE_WARPED_EURC["asset_id"]: BASE_WARPED_EURC,
    ETH_WARPED_USDC["asset_id"]: ETH_WARPED_USDC,
    ETH_WARPED_MILLIETH["asset_id"]: ETH_WARPED_MILLIETH,
    ETH_WARPED_USDT["asset_id"]: ETH_WARPED_USDT,
}



from __future__ import annotations

import json
import logging
from sqlite3 import Row
from typing import Optional, TypeVar, Union

from klvm.casts import int_from_bytes

from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend
from chik.types.condition_opcodes import ConditionOpcode
from chik.util.condition_tools import conditions_dict_for_solution
from chik.util.db_wrapper import DBWrapper2, execute_fetchone
from chik.util.ints import uint32, uint64
from chik.wallet import singleton
from chik.wallet.lineage_proof import LineageProof
from chik.wallet.singleton import get_inner_puzzle_from_singleton, get_singleton_id_from_puzzle
from chik.wallet.singleton_record import SingletonRecord

log = logging.getLogger(__name__)
_T_WalletSingletonStore = TypeVar("_T_WalletSingletonStore", bound="WalletSingletonStore")


class WalletSingletonStore:
    db_wrapper: DBWrapper2

    @classmethod
    async def create(cls: type[_T_WalletSingletonStore], wrapper: DBWrapper2) -> _T_WalletSingletonStore:
        self = cls()
        self.db_wrapper = wrapper

        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await conn.execute(
                "CREATE TABLE IF NOT EXISTS singletons("
                "coin_id blob PRIMARY KEY,"
                " coin text,"
                " singleton_id blob,"
                " wallet_id int,"
                " parent_coin_spend blob,"
                " inner_puzzle_hash blob,"
                " pending tinyint,"
                " removed_height int,"
                " lineage_proof blob,"
                " custom_data blob)"
            )

            await conn.execute("CREATE INDEX IF NOT EXISTS removed_height_index on singletons(removed_height)")

        return self

    async def save_singleton(self, record: SingletonRecord) -> None:
        singleton_id = singleton.get_singleton_id_from_puzzle(record.parent_coinspend.puzzle_reveal)
        if singleton_id is None:  # pragma: no cover
            raise RuntimeError(
                "Failed to derive Singleton ID from puzzle reveal in parent spend %s", record.parent_coinspend
            )
        pending_int = 0
        if record.pending:
            pending_int = 1
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            columns = (
                "coin_id, coin, singleton_id, wallet_id, parent_coin_spend, inner_puzzle_hash, "
                "pending, removed_height, lineage_proof, custom_data"
            )
            await conn.execute(
                f"INSERT or REPLACE INTO singletons ({columns}) VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    record.coin.name().hex(),
                    json.dumps(record.coin.to_json_dict()),
                    singleton_id.hex(),
                    record.wallet_id,
                    bytes(record.parent_coinspend),
                    record.inner_puzzle_hash,
                    pending_int,
                    record.removed_height,
                    bytes(record.lineage_proof),
                    record.custom_data,
                ),
            )

    async def add_spend(
        self,
        wallet_id: uint32,
        coin_state: CoinSpend,
        block_height: uint32 = uint32(0),
        pending: bool = True,
    ) -> None:
        """Given a coin spend of a singleton, attempt to calculate the child coin and details
        for the new singleton record. Add the new record to the store and remove the old record
        if it exists
        """
        # get singleton_id from puzzle_reveal
        singleton_id = get_singleton_id_from_puzzle(coin_state.puzzle_reveal)
        if not singleton_id:
            raise RuntimeError("Coin to add is not a valid singleton")

        # get details for singleton record
        conditions = conditions_dict_for_solution(
            coin_state.puzzle_reveal, coin_state.solution, DEFAULT_CONSTANTS.MAX_BLOCK_COST_KLVM
        )

        cc_cond = next(
            cond for cond in conditions[ConditionOpcode.CREATE_COIN] if int_from_bytes(cond.vars[1]) % 2 == 1
        )

        coin = Coin(coin_state.coin.name(), cc_cond.vars[0], uint64(int_from_bytes(cc_cond.vars[1])))
        inner_puz = get_inner_puzzle_from_singleton(coin_state.puzzle_reveal)
        if inner_puz is None:  # pragma: no cover
            raise RuntimeError("Could not get inner puzzle from puzzle reveal in coin spend %s", coin_state)

        lineage_bytes = [x.as_atom() for x in coin_state.solution.to_program().first().as_iter()]
        if len(lineage_bytes) == 2:
            lineage_proof = LineageProof(bytes32(lineage_bytes[0]), None, uint64(int_from_bytes(lineage_bytes[1])))
        else:
            lineage_proof = LineageProof(
                bytes32(lineage_bytes[0]), bytes32(lineage_bytes[1]), uint64(int_from_bytes(lineage_bytes[2]))
            )
        # Create and save the new singleton record
        new_record = SingletonRecord(
            coin, singleton_id, wallet_id, coin_state, inner_puz.get_tree_hash(), pending, 0, lineage_proof, None
        )
        await self.save_singleton(new_record)
        # check if coin is in DB and mark deleted if found
        current_records = await self.get_records_by_coin_id(coin_state.coin.name())
        if len(current_records) > 0:
            await self.delete_singleton_by_coin_id(coin_state.coin.name(), block_height)

    def _to_singleton_record(self, row: Row) -> SingletonRecord:
        return SingletonRecord(
            coin=Coin.from_json_dict(json.loads(row[1])),
            singleton_id=bytes32.from_hexstr(row[2]),
            wallet_id=uint32(row[3]),
            parent_coinspend=CoinSpend.from_bytes(row[4]),
            inner_puzzle_hash=bytes32.from_bytes(row[5]),  # inner puz hash
            pending=True if row[6] == 1 else False,
            removed_height=uint32(row[7]),
            lineage_proof=LineageProof.from_bytes(row[8]),
            custom_data=row[9],
        )

    async def delete_singleton_by_singleton_id(self, singleton_id: bytes32, height: uint32) -> bool:
        """Tries to mark a given singleton as deleted at specific height

        This is due to how re-org works
        Returns `True` if singleton was found and marked deleted or `False` if not."""
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            cursor = await conn.execute(
                "UPDATE singletons SET removed_height=? WHERE singleton_id=?", (int(height), singleton_id.hex())
            )
            if cursor.rowcount > 0:
                log.info("Deleted singleton with singleton id: %s", singleton_id.hex())
                return True
            log.warning("Couldn't find singleton with singleton id to delete: %s", singleton_id.hex())
            return False

    async def delete_singleton_by_coin_id(self, coin_id: bytes32, height: uint32) -> bool:
        """Tries to mark a given singleton as deleted at specific height

        This is due to how re-org works
        Returns `True` if singleton was found and marked deleted or `False` if not."""
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            cursor = await conn.execute(
                "UPDATE singletons SET removed_height=? WHERE coin_id=?", (int(height), coin_id.hex())
            )
            if cursor.rowcount > 0:
                log.info("Deleted singleton with coin id: %s", coin_id.hex())
                return True
            log.warning("Couldn't find singleton with coin id to delete: %s", coin_id.hex())
            return False

    async def delete_wallet(self, wallet_id: uint32) -> None:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            cursor = await conn.execute("DELETE FROM singletons WHERE wallet_id=?", (wallet_id,))
            await cursor.close()

    async def update_pending_transaction(self, coin_id: bytes32, pending: bool) -> bool:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            c = await conn.execute(
                "UPDATE singletons SET pending=? WHERE coin_id = ?",
                (pending, coin_id.hex()),
            )
            return c.rowcount > 0

    async def get_records_by_wallet_id(self, wallet_id: int) -> list[SingletonRecord]:
        """
        Retrieves all entries for a wallet ID.
        """

        async with self.db_wrapper.reader_no_transaction() as conn:
            rows = await conn.execute_fetchall(
                "SELECT * FROM singletons WHERE wallet_id = ? ORDER BY removed_height",
                (wallet_id,),
            )
        return [self._to_singleton_record(row) for row in rows]

    async def get_records_by_coin_id(self, coin_id: bytes32) -> list[SingletonRecord]:
        """
        Retrieves all entries for a coin ID.
        """

        async with self.db_wrapper.reader_no_transaction() as conn:
            rows = await conn.execute_fetchall(
                "SELECT * FROM singletons WHERE coin_id = ?",
                (coin_id.hex(),),
            )
        return [self._to_singleton_record(row) for row in rows]

    async def get_records_by_singleton_id(self, singleton_id: bytes32) -> list[SingletonRecord]:
        """
        Retrieves all entries for a singleton ID.
        """

        async with self.db_wrapper.reader_no_transaction() as conn:
            rows = await conn.execute_fetchall(
                "SELECT * FROM singletons WHERE singleton_id = ? ORDER BY removed_height",
                (singleton_id.hex(),),
            )
        return [self._to_singleton_record(row) for row in rows]

    async def rollback(self, height: int, wallet_id_arg: int) -> None:
        """
        Rollback removes all entries which have entry_height > height passed in. Note that this is not committed to the
        DB until db_wrapper.commit() is called. However, it is written to the cache, so it can be fetched with
        get_all_state_transitions.
        """

        async with self.db_wrapper.writer_maybe_transaction() as conn:
            cursor = await conn.execute(
                "DELETE FROM singletons WHERE removed_height>? AND wallet_id=?", (height, wallet_id_arg)
            )
            await cursor.close()

    async def count(self, wallet_id: Optional[uint32] = None) -> int:
        sql = "SELECT COUNT(singleton_id) FROM singletons WHERE removed_height=0"
        params: list[uint32] = []
        if wallet_id is not None:
            sql += " AND wallet_id=?"
            params.append(wallet_id)
        async with self.db_wrapper.reader_no_transaction() as conn:
            count_row = await execute_fetchone(conn, sql, params)
            if count_row:
                return int(count_row[0])
        return -1  # pragma: no cover

    async def is_empty(self, wallet_id: Optional[uint32] = None) -> bool:
        sql = "SELECT 1 FROM singletons WHERE removed_height=0"
        params: list[Union[uint32, bytes32]] = []
        if wallet_id is not None:
            sql += " AND wallet_id=?"
            params.append(wallet_id)
        sql += " LIMIT 1"
        async with self.db_wrapper.reader_no_transaction() as conn:
            count_row = await execute_fetchone(conn, sql, params)
            if count_row:
                return False
        return True



from __future__ import annotations

import logging
import time
from typing import TYPE_CHECKING, ClassVar, Optional, cast

from chik.protocols import timelord_protocol
from chik.protocols.timelord_protocol import NewPeakTimelord
from chik.rpc.rpc_server import StateChangedProtocol
from chik.server.api_protocol import ApiMetadata
from chik.timelord.iters_from_block import iters_from_block
from chik.timelord.timelord import Timelord
from chik.timelord.types import Chain, IterationType
from chik.util.ints import uint64

log = logging.getLogger(__name__)


class TimelordAPI:
    if TYPE_CHECKING:
        from chik.server.api_protocol import ApiProtocol

        _protocol_check: ClassVar[ApiProtocol] = cast("TimelordAPI", None)

    log: logging.Logger
    timelord: Timelord
    metadata: ClassVar[ApiMetadata] = ApiMetadata()

    def __init__(self, timelord) -> None:
        self.log = logging.getLogger(__name__)
        self.timelord = timelord

    def ready(self) -> bool:
        return True

    def _set_state_changed_callback(self, callback: StateChangedProtocol) -> None:
        self.timelord.state_changed_callback = callback

    @metadata.request()
    async def new_peak_timelord(self, new_peak: timelord_protocol.NewPeakTimelord) -> None:
        if self.timelord.last_state is None:
            return None
        async with self.timelord.lock:
            if self.timelord.bluebox_mode:
                return None
            self.timelord.max_allowed_inactivity_time = 60

            if self.timelord.last_state.peak is None:
                # no known peak
                log.info("no last known peak, switching to new peak")
                self.timelord.new_peak = new_peak
                self.timelord.state_changed("new_peak", {"height": new_peak.reward_chain_block.height})
                return

            if self.timelord.last_state.get_weight() < new_peak.reward_chain_block.weight:
                # if there is an unfinished block with less iterations, skip so we dont orphan it
                if (
                    new_peak.reward_chain_block.height == self.timelord.last_state.last_height + 1
                    and self.check_orphaned_unfinished_block(new_peak) is True
                ):
                    log.info("there is an unfinished block that this peak would orphan - skip peak")
                    self.timelord.state_changed("skipping_peak", {"height": new_peak.reward_chain_block.height})
                    return

                log.info("Not skipping peak, don't have. Maybe we are not the fastest timelord")
                log.info(
                    f"New peak: height: {new_peak.reward_chain_block.height} weight: "
                    f"{new_peak.reward_chain_block.weight} "
                )
                self.timelord.new_peak = new_peak
                self.timelord.state_changed("new_peak", {"height": new_peak.reward_chain_block.height})
                return

            if self.timelord.last_state.peak.reward_chain_block.get_hash() == new_peak.reward_chain_block.get_hash():
                log.info("Skipping peak, already have.")
            else:
                log.info("Skipping peak, block has equal or lower weight then our peak.")
                log.debug(
                    f"new peak height {new_peak.reward_chain_block.height} "
                    f"weight {new_peak.reward_chain_block.weight}"
                )

            self.timelord.state_changed("skipping_peak", {"height": new_peak.reward_chain_block.height})

    def check_orphaned_unfinished_block(self, new_peak: NewPeakTimelord):
        for unf_block in self.timelord.unfinished_blocks:
            if unf_block.reward_chain_block.total_iters <= new_peak.reward_chain_block.total_iters:
                # there is an unfinished block that would be orphaned by this peak
                return True
        for unf_block in self.timelord.overflow_blocks:
            if unf_block.reward_chain_block.total_iters <= new_peak.reward_chain_block.total_iters:
                # there is an unfinished block (overflow) that would be orphaned by this peak
                return True
        return False

    @metadata.request()
    async def new_unfinished_block_timelord(self, new_unfinished_block: timelord_protocol.NewUnfinishedBlockTimelord):
        if self.timelord.last_state is None:
            return None
        async with self.timelord.lock:
            if self.timelord.bluebox_mode:
                return None
            try:
                sp_iters, ip_iters = iters_from_block(
                    self.timelord.constants,
                    new_unfinished_block.reward_chain_block,
                    self.timelord.last_state.get_sub_slot_iters(),
                    self.timelord.last_state.get_difficulty(),
                    self.timelord.get_height(),
                )
            except Exception:
                return None
            last_ip_iters = self.timelord.last_state.get_last_ip()
            if sp_iters > ip_iters:
                self.timelord.overflow_blocks.append(new_unfinished_block)
                log.debug(f"Overflow unfinished block, total {self.timelord.total_unfinished}")
            elif ip_iters > last_ip_iters:
                new_block_iters: Optional[uint64] = self.timelord._can_infuse_unfinished_block(new_unfinished_block)
                if new_block_iters:
                    self.timelord.unfinished_blocks.append(new_unfinished_block)
                    for chain in [Chain.REWARD_CHAIN, Chain.CHALLENGE_CHAIN]:
                        self.timelord.iters_to_submit[chain].append(new_block_iters)
                    if self.timelord.last_state.get_deficit() < self.timelord.constants.MIN_BLOCKS_PER_CHALLENGE_BLOCK:
                        self.timelord.iters_to_submit[Chain.INFUSED_CHALLENGE_CHAIN].append(new_block_iters)
                    self.timelord.iteration_to_proof_type[new_block_iters] = IterationType.INFUSION_POINT
                    self.timelord.total_unfinished += 1
                    log.debug(f"Non-overflow unfinished block, total {self.timelord.total_unfinished}")

    @metadata.request()
    async def request_compact_proof_of_time(self, vdf_info: timelord_protocol.RequestCompactProofOfTime):
        async with self.timelord.lock:
            if not self.timelord.bluebox_mode:
                return None
            now = time.time()
            # work older than 5s can safely be assumed to be from the previous batch, and needs to be cleared
            while self.timelord.pending_bluebox_info and (now - self.timelord.pending_bluebox_info[0][0] > 5):
                del self.timelord.pending_bluebox_info[0]
            self.timelord.pending_bluebox_info.append((now, vdf_info))



from __future__ import annotations

import pytest

from chik.types.aliases import TimelordService


@pytest.mark.anyio
async def test_timelord_has_no_server(timelord_service: TimelordService) -> None:
    timelord_server = timelord_service._node.server
    assert timelord_server.webserver is None



from __future__ import annotations

import sys
from collections.abc import AsyncGenerator
from pathlib import Path
from typing import Any, Optional

from chik_rs import PrivateKey

from chik.consensus.coinbase import create_puzzlehash_for_pk
from chik.daemon.server import WebSocketServer, daemon_launch_lock_path
from chik.server.signal_handlers import SignalHandlers
from chik.simulator.full_node_simulator import FullNodeSimulator
from chik.simulator.socket import find_available_listen_port
from chik.simulator.ssl_certs import (
    SSLTestCACertAndPrivateKey,
    SSLTestCollateralWrapper,
    SSLTestNodeCertsAndKeys,
    get_next_nodes_certs_and_keys,
    get_next_private_ca_cert_and_key,
)
from chik.simulator.start_simulator import async_main as start_simulator_main
from chik.ssl.create_ssl import create_all_ssl
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.bech32m import encode_puzzle_hash
from chik.util.config import create_default_chik_config, load_config, save_config
from chik.util.errors import KeychainFingerprintExists
from chik.util.ints import uint32
from chik.util.keychain import Keychain
from chik.util.lock import Lockfile
from chik.wallet.derive_keys import master_sk_to_wallet_sk

"""
These functions are used to test the simulator.
"""


def mnemonic_fingerprint(keychain: Keychain) -> tuple[str, int]:
    mnemonic = (
        "today grape album ticket joy idle supreme sausage "
        "oppose voice angle roast you oven betray exact "
        "memory riot escape high dragon knock food blade"
    )
    # add key to keychain
    try:
        sk = keychain.add_key(mnemonic)
    except KeychainFingerprintExists:
        pass
    fingerprint = sk.get_g1().get_fingerprint()
    return mnemonic, fingerprint


def get_puzzle_hash_from_key(keychain: Keychain, fingerprint: int, key_id: int = 1) -> bytes32:
    priv_key_and_entropy = keychain.get_private_key_by_fingerprint(fingerprint)
    if priv_key_and_entropy is None:
        raise Exception("Fingerprint not found")
    private_key = priv_key_and_entropy[0]
    sk_for_wallet_id: PrivateKey = master_sk_to_wallet_sk(private_key, uint32(key_id))
    puzzle_hash: bytes32 = create_puzzlehash_for_pk(sk_for_wallet_id.get_g1())
    return puzzle_hash


def create_config(
    chik_root: Path,
    fingerprint: int,
    private_ca_crt_and_key: tuple[bytes, bytes],
    node_certs_and_keys: dict[str, dict[str, dict[str, bytes]]],
    keychain: Keychain,
) -> dict[str, Any]:
    # create chik directories
    create_default_chik_config(chik_root)
    create_all_ssl(
        chik_root,
        private_ca_crt_and_key=private_ca_crt_and_key,
        node_certs_and_keys=node_certs_and_keys,
    )
    # load config
    config = load_config(chik_root, "config.yaml")
    config["full_node"]["send_uncompact_interval"] = 0
    config["full_node"]["target_uncompact_proofs"] = 30
    config["full_node"]["peer_connect_interval"] = 50
    config["full_node"]["sanitize_weight_proof_only"] = False
    config["full_node"]["introducer_peer"] = None
    config["full_node"]["dns_servers"] = []
    config["logging"]["log_stdout"] = True
    config["selected_network"] = "testnet0"
    for service in [
        "harvester",
        "farmer",
        "full_node",
        "wallet",
        "introducer",
        "timelord",
        "pool",
        "simulator",
    ]:
        config[service]["selected_network"] = "testnet0"
    config["daemon_port"] = find_available_listen_port("BlockTools daemon")
    config["full_node"]["port"] = 0
    config["full_node"]["rpc_port"] = find_available_listen_port("Node RPC")
    # simulator overrides
    config["simulator"]["key_fingerprint"] = fingerprint
    config["simulator"]["farming_address"] = encode_puzzle_hash(get_puzzle_hash_from_key(keychain, fingerprint), "txck")
    config["simulator"]["plot_directory"] = "test-simulator/plots"
    # save config
    save_config(chik_root, "config.yaml", config)
    return config


async def start_simulator(chik_root: Path, automated_testing: bool = False) -> AsyncGenerator[FullNodeSimulator, None]:
    sys.argv = [sys.argv[0]]  # clear sys.argv to avoid issues with config.yaml
    started_simulator = await start_simulator_main(True, automated_testing, root_path=chik_root)
    service = started_simulator.service

    async with service.manage():
        yield service._api


async def get_full_chik_simulator(
    chik_root: Path,
    keychain: Optional[Keychain] = None,
    automated_testing: bool = False,
    config: Optional[dict[str, Any]] = None,
) -> AsyncGenerator[tuple[FullNodeSimulator, Path, dict[str, Any], str, int, Keychain], None]:
    """
    A chik root Path is required.
    The chik root Path can be a temporary directory (tempfile.TemporaryDirectory)
    Passing in a Keychain prevents test keys from being added to the default key location
    This test can either be run in automated mode or not, which determines which mode block tools run in.
    This test is fully interdependent and can be used without the rest of the chik test suite.
    Please refer to the documentation for more information.
    """

    if keychain is None:
        keychain = Keychain()

    with Lockfile.create(daemon_launch_lock_path(chik_root)):
        mnemonic, fingerprint = mnemonic_fingerprint(keychain)

        ssl_ca_cert_and_key_wrapper: SSLTestCollateralWrapper[SSLTestCACertAndPrivateKey] = (
            get_next_private_ca_cert_and_key()
        )
        ssl_nodes_certs_and_keys_wrapper: SSLTestCollateralWrapper[SSLTestNodeCertsAndKeys] = (
            get_next_nodes_certs_and_keys()
        )
        if config is None:
            config = create_config(
                chik_root,
                fingerprint,
                ssl_ca_cert_and_key_wrapper.collateral.cert_and_key,
                ssl_nodes_certs_and_keys_wrapper.collateral.certs_and_keys,
                keychain,
            )
        crt_path = chik_root / config["daemon_ssl"]["private_crt"]
        key_path = chik_root / config["daemon_ssl"]["private_key"]
        ca_crt_path = chik_root / config["private_ssl_ca"]["crt"]
        ca_key_path = chik_root / config["private_ssl_ca"]["key"]

        ws_server = WebSocketServer(chik_root, ca_crt_path, ca_key_path, crt_path, key_path)
        async with SignalHandlers.manage() as signal_handlers:
            await ws_server.setup_process_global_state(signal_handlers=signal_handlers)
            async with ws_server.run():
                async for simulator in start_simulator(chik_root, automated_testing):
                    yield simulator, chik_root, config, mnemonic, fingerprint, keychain



from __future__ import annotations

import random
import sqlite3
from contextlib import closing
from pathlib import Path

import pytest

from chik._tests.util.temp_file import TempFile
from chik.cmds.db_validate_func import validate_v2
from chik.consensus.block_body_validation import ForkInfo
from chik.consensus.blockchain import Blockchain
from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.consensus.multiprocess_validation import PreValidationResult
from chik.full_node.block_store import BlockStore
from chik.full_node.coin_store import CoinStore
from chik.simulator.block_tools import test_constants
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.full_block import FullBlock
from chik.util.db_wrapper import DBWrapper2
from chik.util.ints import uint32, uint64


def rand_hash() -> bytes32:
    ret = bytearray(32)
    for i in range(32):
        ret[i] = random.getrandbits(8)
    return bytes32(ret)


def make_version(conn: sqlite3.Connection, version: int) -> None:
    conn.execute("CREATE TABLE database_version(version int)")
    conn.execute("INSERT INTO database_version VALUES (?)", (version,))
    conn.commit()


def make_peak(conn: sqlite3.Connection, peak_hash: bytes32) -> None:
    conn.execute("CREATE TABLE IF NOT EXISTS current_peak(key int PRIMARY KEY, hash blob)")
    conn.execute("INSERT OR REPLACE INTO current_peak VALUES(?, ?)", (0, peak_hash))
    conn.commit()


def make_block_table(conn: sqlite3.Connection) -> None:
    conn.execute(
        "CREATE TABLE IF NOT EXISTS full_blocks("
        "header_hash blob PRIMARY KEY,"
        "prev_hash blob,"
        "height bigint,"
        "sub_epoch_summary blob,"
        "is_fully_compactified tinyint,"
        "in_main_chain tinyint,"
        "block blob,"
        "block_record blob)"
    )


def add_block(
    conn: sqlite3.Connection, header_hash: bytes32, prev_hash: bytes32, height: int, in_main_chain: bool
) -> None:
    conn.execute(
        "INSERT INTO full_blocks VALUES(?, ?, ?, NULL, 0, ?, NULL, NULL)",
        (
            header_hash,
            prev_hash,
            height,
            in_main_chain,
        ),
    )


def test_db_validate_wrong_version() -> None:
    with TempFile() as db_file:
        with closing(sqlite3.connect(db_file)) as conn:
            make_version(conn, 3)

        with pytest.raises(RuntimeError) as execinfo:
            validate_v2(db_file, validate_blocks=False)
        assert "Database has the wrong version (3 expected 2)" in str(execinfo.value)


def test_db_validate_missing_peak_table() -> None:
    with TempFile() as db_file:
        with closing(sqlite3.connect(db_file)) as conn:
            make_version(conn, 2)

        with pytest.raises(RuntimeError) as execinfo:
            validate_v2(db_file, validate_blocks=False)
        assert "Database is missing current_peak table" in str(execinfo.value)


def test_db_validate_missing_peak_block() -> None:
    with TempFile() as db_file:
        with closing(sqlite3.connect(db_file)) as conn:
            make_version(conn, 2)
            make_peak(conn, bytes32.fromhex("fafafafafafafafafafafafafafafafafafafafafafafafafafafafafafafafa"))

            make_block_table(conn)

        with pytest.raises(RuntimeError) as execinfo:
            validate_v2(db_file, validate_blocks=False)
        assert "Database is missing the peak block" in str(execinfo.value)


@pytest.mark.parametrize("invalid_in_chain", [True, False])
def test_db_validate_in_main_chain(invalid_in_chain: bool) -> None:
    with TempFile() as db_file:
        with closing(sqlite3.connect(db_file)) as conn:
            make_version(conn, 2)
            make_block_table(conn)

            prev = bytes32(DEFAULT_CONSTANTS.AGG_SIG_ME_ADDITIONAL_DATA)
            for height in range(0, 100):
                header_hash = rand_hash()
                add_block(conn, header_hash, prev, height, True)
                if height % 4 == 0:
                    # insert an orphaned block
                    add_block(conn, rand_hash(), prev, height, invalid_in_chain)
                prev = header_hash

            make_peak(conn, header_hash)

        if invalid_in_chain:
            with pytest.raises(RuntimeError) as execinfo:
                validate_v2(db_file, validate_blocks=False)
            assert " (height: 96) is orphaned, but in_main_chain is set" in str(execinfo.value)
        else:
            validate_v2(db_file, validate_blocks=False)


async def make_db(db_file: Path, blocks: list[FullBlock]) -> None:
    async with DBWrapper2.managed(database=db_file, reader_count=1, db_version=2) as db_wrapper:
        async with db_wrapper.writer_maybe_transaction() as conn:
            # this is done by chik init normally
            await conn.execute("CREATE TABLE database_version(version int)")
            await conn.execute("INSERT INTO database_version VALUES (2)")

        block_store = await BlockStore.create(db_wrapper)
        coin_store = await CoinStore.create(db_wrapper)

        bc = await Blockchain.create(coin_store, block_store, test_constants, Path("."), reserved_cores=0)
        sub_slot_iters = test_constants.SUB_SLOT_ITERS_STARTING
        for block in blocks:
            if block.height != 0 and len(block.finished_sub_slots) > 0:
                if block.finished_sub_slots[0].challenge_chain.new_sub_slot_iters is not None:
                    sub_slot_iters = block.finished_sub_slots[0].challenge_chain.new_sub_slot_iters
            results = PreValidationResult(None, uint64(1), None, uint32(0))
            fork_info = ForkInfo(block.height - 1, block.height - 1, block.prev_header_hash)
            _, err, _ = await bc.add_block(block, results, sub_slot_iters=sub_slot_iters, fork_info=fork_info)
            assert err is None


@pytest.mark.anyio
async def test_db_validate_default_1000_blocks(default_1000_blocks: list[FullBlock]) -> None:
    with TempFile() as db_file:
        await make_db(db_file, default_1000_blocks)

        # we expect everything to be valid except this is a test chain, so it
        # doesn't have the correct genesis challenge
        with pytest.raises(RuntimeError) as execinfo:
            validate_v2(db_file, validate_blocks=True)
        assert "Blockchain has invalid genesis challenge" in str(execinfo.value)



from __future__ import annotations

import math
import time
from dataclasses import dataclass

from chik.util.ints import uint32, uint64
from chik.util.streamable import Streamable, streamable


@streamable
@dataclass(frozen=True)
class PeerRecord(Streamable):
    peer_id: str
    ip_address: str
    port: uint32
    connected: bool
    last_try_timestamp: uint64
    try_count: uint32
    connected_timestamp: uint64
    added_timestamp: uint64
    best_timestamp: uint64
    version: str
    handshake_time: uint64
    tls_version: str

    def update_version(self, version: str, now: uint64) -> None:
        if version != "undefined":
            object.__setattr__(self, "version", version)
        object.__setattr__(self, "handshake_time", now)


@dataclass
class PeerStat:
    weight: float
    count: float
    reliability: float

    def update(self, is_reachable: bool, age: int, tau: int) -> None:
        f = math.exp(-age / tau)
        self.reliability = self.reliability * f + (1.0 - f if is_reachable else 0.0)
        self.count = self.count * f + 1.0
        self.weight = self.weight * f + 1.0 - f


class PeerReliability:
    peer_id: str
    ignore_till: int
    ban_till: int
    stat_2h: PeerStat
    stat_8h: PeerStat
    stat_1d: PeerStat
    stat_1w: PeerStat
    stat_1m: PeerStat
    tries: int
    successes: int

    def __init__(
        self,
        peer_id: str,
        ignore_till: int = 0,
        ban_till: int = 0,
        stat_2h_weight: float = 0.0,
        stat_2h_count: float = 0.0,
        stat_2h_reliability: float = 0.0,
        stat_8h_weight: float = 0.0,
        stat_8h_count: float = 0.0,
        stat_8h_reliability: float = 0.0,
        stat_1d_weight: float = 0.0,
        stat_1d_count: float = 0.0,
        stat_1d_reliability: float = 0.0,
        stat_1w_weight: float = 0.0,
        stat_1w_count: float = 0.0,
        stat_1w_reliability: float = 0.0,
        stat_1m_weight: float = 0.0,
        stat_1m_count: float = 0.0,
        stat_1m_reliability: float = 0.0,
        tries: int = 0,
        successes: int = 0,
    ) -> None:
        self.peer_id = peer_id
        self.ignore_till = ignore_till
        self.ban_till = ban_till
        self.stat_2h = PeerStat(stat_2h_weight, stat_2h_count, stat_2h_reliability)
        self.stat_8h = PeerStat(stat_8h_weight, stat_8h_count, stat_8h_reliability)
        self.stat_1d = PeerStat(stat_1d_weight, stat_1d_count, stat_1d_reliability)
        self.stat_1w = PeerStat(stat_1w_weight, stat_1w_count, stat_1w_reliability)
        self.stat_1m = PeerStat(stat_1m_weight, stat_1m_count, stat_1m_reliability)
        self.tries = tries
        self.successes = successes

    def is_reliable(self) -> bool:
        if self.tries > 0 and self.tries <= 3 and self.successes * 2 >= self.tries:
            return True
        if self.stat_2h.reliability > 0.85 and self.stat_2h.count > 2:
            return True
        if self.stat_8h.reliability > 0.7 and self.stat_8h.count > 4:
            return True
        if self.stat_1d.reliability > 0.55 and self.stat_1d.count > 8:
            return True
        if self.stat_1w.reliability > 0.45 and self.stat_1w.count > 16:
            return True
        if self.stat_1m.reliability > 0.35 and self.stat_1m.count > 32:
            return True
        return False

    def get_ban_time(self) -> int:
        if self.is_reliable():
            return 0
        if self.stat_1m.reliability - self.stat_1m.weight + 1 < 0.15 and self.stat_1m.count > 32:
            return 30 * 86400
        if self.stat_1w.reliability - self.stat_1w.weight + 1.0 < 0.10 and self.stat_1w.count > 16:
            return 7 * 86400
        if self.stat_1d.reliability - self.stat_1d.weight + 1.0 < 0.05 and self.stat_1d.count > 8:
            return 86400
        return 0

    def get_ignore_time(self) -> int:
        if self.is_reliable():
            return 0
        if self.stat_1m.reliability - self.stat_1m.weight + 1.0 < 0.20 and self.stat_1m.count > 2:
            return 10 * 86400
        if self.stat_1w.reliability - self.stat_1w.weight + 1.0 < 0.16 and self.stat_1w.count > 2:
            return 3 * 86400
        if self.stat_1d.reliability - self.stat_1d.weight + 1.0 < 0.12 and self.stat_1d.count > 2:
            return 8 * 3600
        if self.stat_8h.reliability - self.stat_8h.weight + 1.0 < 0.08 and self.stat_8h.count > 2:
            return 2 * 3600
        return 0

    def update(self, is_reachable: bool, age: int) -> None:
        self.stat_2h.update(is_reachable, age, 2 * 3600)
        self.stat_8h.update(is_reachable, age, 8 * 3600)
        self.stat_1d.update(is_reachable, age, 24 * 3600)
        self.stat_1w.update(is_reachable, age, 7 * 24 * 3600)
        self.stat_1m.update(is_reachable, age, 24 * 30 * 3600)
        self.tries += 1
        if is_reachable:
            self.successes += 1
        current_ignore_time = self.get_ignore_time()
        now = int(time.time())
        if current_ignore_time > 0 and (self.ignore_till == 0 or self.ignore_till < current_ignore_time + now):
            self.ignore_till = current_ignore_time + now
        current_ban_time = self.get_ban_time()
        if current_ban_time > 0 and (self.ban_till == 0 or self.ban_till < current_ban_time + now):
            self.ban_till = current_ban_time + now
